# Copyright (c) 2014 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
from oslo_log import log as logging
from pecan import hooks

from blazar.api import context
from blazar.db import api as dbapi
from blazar.manager.leases import rpcapi as leases_rpcapi
from blazar.manager.oshosts import rpcapi as hosts_rpcapi

LOG = logging.getLogger(__name__)


class ConfigHook(hooks.PecanHook):
    """ConfigHook

    Attach the configuration object to the request
    so controllers can get to it.
    """

    def before(self, state):
        state.request.cfg = cfg.CONF


class DBHook(hooks.PecanHook):
    """Attach the dbapi object to the request so controllers can get to it."""

    def before(self, state):
        state.request.dbapi = dbapi.get_instance()


class ContextHook(hooks.PecanHook):
    """Configures a request context and attaches it to the request."""

    def before(self, state):
        state.request.context = context.ctx_from_headers(state.request.headers)
        state.request.context.__enter__()

    # NOTE(sbauza): on_error() can be fired before after() if the original
    #               exception is not catched by WSME. That's necessary to not
    #               handle context.__exit__() within on_error() as it could
    #               lead to pop the stack twice for the same request
    def after(self, state):
        # If no API extensions are loaded, context is empty
        if state.request.context:
            state.request.context.__exit__(None, None, None)


class RPCHook(hooks.PecanHook):
    """Attach the rpcapi object to the request so controllers can get to it."""

    def before(self, state):
        state.request.rpcapi = leases_rpcapi.ManagerRPCAPI()
        state.request.hosts_rpcapi = hosts_rpcapi.ManagerRPCAPI()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from blazar import exceptions
from blazar.i18n import _


class ResourceProviderRetrievalFailed(exceptions.BlazarException):
    msg_fmt = _("Failed to get resource provider %(name)s")


class ResourceProviderCreationFailed(exceptions.BlazarException):
    msg_fmt = _("Failed to create resource provider %(name)s")


class ResourceProviderCreationConflict(exceptions.BlazarException):
    code = 409
    msg_fmt = _("Conflict on creating resource provider %(name)s ")


class ResourceProviderDeletionFailed(exceptions.BlazarException):
    msg_fmt = _("Failed to delete resource provider %(uuid)s")


class ResourceClassCreationFailed(exceptions.BlazarException):
    msg_fmt = _("Failed to create resource class '%(resource_class)s'")


class ResourceClassDeletionFailed(exceptions.BlazarException):
    msg_fmt = _("Failed to delete resource class '%(resource_class)s'")


class ResourceProviderNotFound(exceptions.BlazarException):
    msg_fmt = _("No such resource provider %(resource_provider)s")


class InventoryUpdateFailed(exceptions.BlazarException):
    msg_fmt = _("Failed to update the inventory of resource provider "
                "%(resource_provider)s")


class InventoryConflict(exceptions.BlazarException):
    code = 409
    msg_fmt = _("Conflict on updating inventory on resource provider "
                "%(resource_provider)s")


class FloatingIPNetworkNotFound(exceptions.InvalidInput):
    msg_fmt = _("Failed to find network %(network)s")


class FloatingIPSubnetNotFound(exceptions.NotFound):
    msg_fmt = _("Valid subnet for the floating IP %(fip)s is not found.")


class NeutronUsesFloatingIP(exceptions.InvalidInput):
    msg_fmt = _("The floating IP %(floatingip)s is used in allocation_pools "
                "or gateway_ip in subnet %(subnet)s .")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_service import threadgroup

from blazar.db import api as db_api
from blazar import exceptions
from blazar.monitor import base as base_monitor
from blazar.plugins import base
from blazar import tests


HEALING_INTERVAL = 10


class DummyMonitorPlugin(base.BaseMonitorPlugin):
    def is_notification_enabled(self):
        return True

    def get_notification_event_types(self):
        return []

    def get_notification_topics(self):
        return []

    def notification_callback(self, event_type, message):
        return {}

    def is_polling_enabled(self):
        return False

    def get_polling_interval(self):
        return 0

    def poll(self):
        return {}

    def get_healing_interval(self):
        return HEALING_INTERVAL

    def heal(self):
        return {}


class BaseMonitorTestCase(tests.TestCase):
    def setUp(self):
        super(BaseMonitorTestCase, self).setUp()
        self.monitor_plugins = [DummyMonitorPlugin()]
        self.monitor = base_monitor.BaseMonitor(self.monitor_plugins)

    def test_start_periodic_healing(self):
        add_timer = self.patch(threadgroup.ThreadGroup, 'add_timer')

        self.monitor.start_periodic_healing()
        add_timer.assert_called_once_with(
            HEALING_INTERVAL * 60, self.monitor.call_monitor_plugin, None,
            self.monitor_plugins[0].heal)

    def test_stop_periodic_healing(self):
        dummy_timer = mock.Mock()
        timer_done = self.patch(threadgroup.ThreadGroup, 'timer_done')
        self.monitor.healing_timers.append(dummy_timer)

        self.monitor.stop_monitoring()
        timer_done.assert_called_once_with(dummy_timer)

    def test_call_monitor_plugin(self):
        callback = self.patch(DummyMonitorPlugin,
                              'notification_callback')
        callback.return_value = {
            'dummy_id1': {'missing_resources': True}
        }
        update_flags = self.patch(self.monitor, '_update_flags')

        self.monitor.call_monitor_plugin(callback, 'event_type1', 'hello')
        callback.assert_called_once_with('event_type1', 'hello')
        update_flags.assert_called_once_with(
            {'dummy_id1': {'missing_resources': True}})

    def test_error_in_callback(self):
        callback = self.patch(DummyMonitorPlugin, 'poll')
        callback.side_effect = exceptions.BlazarException('error')

        # Testing that no exception is raised even if the callback raises one
        self.monitor.call_monitor_plugin(callback)

    def test_call_update_flags(self):
        reservation_update = self.patch(db_api, 'reservation_update')
        reservation_get = self.patch(db_api, 'reservation_get')
        reservation_get.return_value = {
            'lease_id': 'dummy_id2'
        }
        lease_update = self.patch(db_api, 'lease_update')

        self.monitor._update_flags({'dummy_id1': {'missing_resources': True}})
        reservation_update.assert_called_once_with(
            'dummy_id1', {'missing_resources': True})
        reservation_get.assert_called_once_with('dummy_id1')
        lease_update.assert_called_once_with('dummy_id2',
                                             {'degraded': True})

    def test_error_in_update_flags(self):
        callback = self.patch(DummyMonitorPlugin, 'poll')
        callback.return_value = {
            'dummy_id1': {'missing_resources': True}
        }

        update_flags = self.patch(self.monitor, '_update_flags')
        update_flags.side_effect = exceptions.BlazarException('error')

        # Testing that no exception is raised even if the callback raises one
        self.monitor.call_monitor_plugin(callback)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log as logging
from oslo_service import threadgroup

from blazar.db import api as db_api

LOG = logging.getLogger(__name__)


class BaseMonitor(object):
    """Base class for monitoring classes."""

    def __init__(self, monitor_plugins):
        self.monitor_plugins = monitor_plugins
        self.tg = threadgroup.ThreadGroup()
        self.healing_timers = []

    def start_monitoring(self):
        """Start monitoring."""
        self.start_periodic_healing()

    def stop_monitoring(self):
        """Stop monitoring."""
        self.stop_periodic_healing()

    def start_periodic_healing(self):
        """Start periodic healing process."""
        for plugin in self.monitor_plugins:
            healing_interval_mins = plugin.get_healing_interval()
            if healing_interval_mins > 0:
                self.healing_timers.append(
                    self.tg.add_timer(healing_interval_mins * 60,
                                      self.call_monitor_plugin,
                                      None,
                                      plugin.heal))

    def stop_periodic_healing(self):
        """Stop periodic healing process."""
        for timer in self.healing_timers:
            self.tg.timer_done(timer)

    def call_monitor_plugin(self, callback, *args, **kwargs):
        """Call a callback and update lease/reservation flags."""
        # This method has to handle any exception internally. It shouldn't
        # raise an exception because the timer threads in the BaseMonitor class
        # terminates its execution once the thread has received any exception.
        try:
            # The callback() has to return a dictionary of
            # {reservation id: flags to update}.
            # e.g. {'dummyid': {'missing_resources': True}}
            reservation_flags = callback(*args, **kwargs)

            if reservation_flags:
                self._update_flags(reservation_flags)
        except Exception as e:
            LOG.exception('Caught an exception while executing a callback. '
                          '%s', str(e))

    def _update_flags(self, reservation_flags):
        """Update lease/reservation flags."""
        lease_ids = set([])

        for reservation_id, flags in reservation_flags.items():
            db_api.reservation_update(reservation_id, flags)
            LOG.debug('Reservation %s was updated: %s',
                      reservation_id, flags)
            reservation = db_api.reservation_get(reservation_id)
            lease_ids.add(reservation['lease_id'])

        for lease_id in lease_ids:
            LOG.debug('Lease %s was updated: {"degraded": True}', lease_id)
            db_api.lease_update(lease_id, {'degraded': True})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy

RULE_ADMIN = 'rule:admin'
RULE_ADMIN_OR_OWNER = 'rule:admin_or_owner'
RULE_ANY = '@'

rules = [
    policy.RuleDefault(
        name="admin",
        check_str="is_admin:True or role:admin",
        description="Default rule for most Admin APIs."),
    policy.RuleDefault(
        name="admin_or_owner",
        check_str="rule:admin or project_id:%(project_id)s",
        description="Default rule for most non-Admin APIs.")
]


def list_rules():
    return rules



# Copyright 2014 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
import oslo_messaging as messaging

from blazar.notification import notifier as notification
from blazar import tests

CONF = cfg.CONF


class FakeNotifier(object):
    def info(self):
        pass


class NotifierTestCase(tests.TestCase):
    def setUp(self):
        super(NotifierTestCase, self).setUp()

        self.group = 'notifications'
        CONF.set_override('publisher_id', 'lease-service', self.group)

        # Fake Oslo notifier
        self.fake_notifier = self.patch(messaging, 'Notifier')
        self.fake_notifier.return_value = FakeNotifier()
        self.fake_transport = self.patch(
            messaging,
            'get_notification_transport')

        self.info_method = self.patch(FakeNotifier, 'info')

        self.context = {'user_id': 1, 'token': 'aabbcc'}
        self.payload = {'id': 1, 'name': 'Lease1', 'start-date': 'now'}

        notification.init()
        self.notifier = notification.Notifier()

    def test_notify_with_wrong_level(self):
        self.notifier._notify(self.context, 'wrong', 'event', self.payload)
        self.info_method.assert_called_once_with(self.context,
                                                 'event', self.payload)

    def test_send_lease_event(self):
        self.notifier.send_lease_notification(self.context, self.payload,
                                              'start')
        self.info_method.assert_called_once_with(self.context,
                                                 'start',
                                                 self.payload)

    def test_cleanup(self):
        notification.cleanup()

        self.fake_transport.return_value.cleanup.assert_called_once_with()
        self.assertIsNone(notification.NOTIFIER)
        self.assertIsNone(notification.TRANSPORT)

    def test_init(self):
        self.fake_transport.assert_called_once_with(notification.CONF)
        self.fake_notifier.assert_called_once_with(
            self.fake_transport.return_value, publisher_id='lease-service')

    def test_init_called_twice_returns_same_instance(self):
        prev_notifier = notification.NOTIFIER
        prev_transport = notification.TRANSPORT

        notification.init()
        self.assertIs(prev_notifier, notification.NOTIFIER)
        self.assertIs(prev_transport, notification.TRANSPORT)



# Copyright (c) 2013 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

from blazar import manager
from blazar.utils import service

CONF = cfg.CONF
CONF.import_opt('rpc_topic', 'blazar.manager.service', 'manager')


class ManagerRPCAPI(service.RPCClient):
    """Client side for the Manager RPC API.

    Used from other services to communicate with blazar-manager service.
    """
    BASE_RPC_API_VERSION = '1.0'

    def __init__(self):
        """Initiate RPC API client with needed topic and RPC version."""
        super(ManagerRPCAPI, self).__init__(manager.get_target())

    def get_computehost(self, host_id):
        """Get detailed info about some computehost."""
        return self.call('physical:host:get_computehost', host_id=host_id)

    def list_computehosts(self, query=None):
        """List all computehosts."""
        return self.call('physical:host:list_computehosts', query=query)

    def create_computehost(self, host_values):
        """Create computehost with specified parameters."""
        return self.call('physical:host:create_computehost',
                         host_values=host_values)

    def update_computehost(self, host_id, values):
        """Update computehost with passes values dictionary."""
        return self.call('physical:host:update_computehost', host_id=host_id,
                         values=values)

    def delete_computehost(self, host_id):
        """Delete specified computehost."""
        return self.call('physical:host:delete_computehost',
                         host_id=host_id)

    def list_allocations(self, query):
        """List all allocations on all computehosts."""
        return self.call('physical:host:list_allocations', query=query)

    def get_allocations(self, host_id, query):
        """List all allocations on a specified computehost."""
        return self.call('physical:host:get_allocations',
                         host_id=host_id, query=query)

    def list_resource_properties(self, query):
        """List resource properties and possible values for computehosts."""
        return self.call('physical:host:list_resource_properties', query=query)

    def update_resource_property(self, property_name, values):
        """Update resource property for computehost."""
        return self.call('physical:host:update_resource_property',
                         property_name=property_name, values=values)



# Copyright (c) 2017 NTT.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from unittest import mock
import uuid

import ddt
from novaclient import exceptions as nova_exceptions

from blazar import context
from blazar.db import api as db_api
from blazar.db import utils as db_utils
from blazar.manager import exceptions as mgr_exceptions
from blazar.plugins import instances
from blazar.plugins.instances import instance_plugin
from blazar.plugins import oshosts
from blazar import tests
from blazar.utils.openstack import nova
from oslo_config import cfg
from oslo_config import fixture as conf_fixture

CONF = cfg.CONF


@ddt.ddt
class TestVirtualInstancePlugin(tests.TestCase):

    def setUp(self):
        super(TestVirtualInstancePlugin, self).setUp()

    def test_configuration(self):
        self.cfg = self.useFixture(conf_fixture.Config(CONF))
        self.cfg.config(os_admin_username='fake-user')
        self.cfg.config(os_admin_password='fake-passwd')
        self.cfg.config(os_admin_user_domain_name='fake-user-domain')
        self.cfg.config(os_admin_project_name='fake-pj-name')
        self.cfg.config(os_admin_project_domain_name='fake-pj-domain')
        plugin = instance_plugin.VirtualInstancePlugin()
        self.assertEqual("fake-user", plugin.username)
        self.assertEqual("fake-passwd", plugin.password)
        self.assertEqual("fake-user-domain", plugin.user_domain_name)
        self.assertEqual("fake-pj-name", plugin.project_name)
        self.assertEqual("fake-pj-domain", plugin.project_domain_name)

    def get_input_values(self, vcpus, memory, disk, amount, affinity,
                         start, end, lease_id, resource_properties):
        values = {'vcpus': vcpus, 'memory_mb': memory, 'disk_gb': disk,
                  'amount': amount, 'affinity': affinity, 'start_date': start,
                  'end_date': end, 'lease_id': lease_id,
                  'resource_properties': resource_properties}
        return values

    def generate_host_info(self, id, vcpus, memory, disk):
        return {'id': id, 'vcpus': vcpus,
                'memory_mb': memory, 'local_gb': disk}

    def generate_event(self, id, lease_id, event_type, time, status='UNDONE'):
        return {
            'id': id,
            'lease_id': lease_id,
            'event_type': event_type,
            'time': time,
            'status': status
            }

    def get_uuid(self):
        return str(uuid.uuid4())

    def generate_basic_events(self, lease_id, start, before_end, end):
        return [
            self.generate_event(self.get_uuid(), lease_id, 'start_lease',
                                datetime.datetime.strptime(start,
                                                           '%Y-%m-%d %H:%M')),
            self.generate_event(self.get_uuid(), lease_id, 'before_end_lease',
                                datetime.datetime.strptime(before_end,
                                                           '%Y-%m-%d %H:%M')),
            self.generate_event(self.get_uuid(), lease_id, 'end_lease',
                                datetime.datetime.strptime(end,
                                                           '%Y-%m-%d %H:%M')),
            ]

    def test_reserve_resource(self):
        plugin = instance_plugin.VirtualInstancePlugin()
        mock_pickup_hosts = self.patch(plugin, 'pickup_hosts')
        mock_pickup_hosts.return_value = {'added': ['host1', 'host2'],
                                          'removed': []}

        mock_inst_create = self.patch(db_api, 'instance_reservation_create')
        fake_instance_reservation = {'id': 'instance-reservation-id1'}
        mock_inst_create.return_value = fake_instance_reservation

        mock_alloc_create = self.patch(db_api, 'host_allocation_create')

        mock_create_resources = self.patch(plugin, '_create_resources')
        mock_flavor = mock.MagicMock(id=1)
        mock_group = mock.MagicMock(id=2)
        mock_pool = mock.MagicMock(id=3)
        mock_create_resources.return_value = (mock_flavor,
                                              mock_group, mock_pool)

        mock_inst_update = self.patch(db_api, 'instance_reservation_update')

        inputs = self.get_input_values(2, 4018, 10, 1, False,
                                       '2030-01-01 08:00', '2030-01-01 08:00',
                                       'lease-1', '')

        expected_ret = 'instance-reservation-id1'

        ret = plugin.reserve_resource('res_id1', inputs)

        self.assertEqual(expected_ret, ret)
        pickup_hosts_value = {}
        for key in ['vcpus', 'memory_mb', 'disk_gb', 'amount', 'affinity',
                    'lease_id', 'start_date', 'end_date',
                    'resource_properties']:
            pickup_hosts_value[key] = inputs[key]
        mock_pickup_hosts.assert_called_once_with('res_id1',
                                                  pickup_hosts_value)

        mock_alloc_create.assert_any_call({'compute_host_id': 'host1',
                                           'reservation_id': 'res_id1'})
        mock_alloc_create.assert_any_call({'compute_host_id': 'host2',
                                           'reservation_id': 'res_id1'})
        mock_create_resources.assert_called_once_with(
            fake_instance_reservation)
        mock_inst_update.assert_called_once_with('instance-reservation-id1',
                                                 {'flavor_id': 1,
                                                  'server_group_id': 2,
                                                  'aggregate_id': 3})

    @ddt.data("abc", 2, "2")
    def test_affinity_error(self, value):
        plugin = instance_plugin.VirtualInstancePlugin()
        inputs = self.get_input_values(2, 4018, 10, 1, value,
                                       '2030-01-01 08:00', '2030-01-01 08:00',
                                       'lease-1', '')
        self.assertRaises(mgr_exceptions.MalformedParameter,
                          plugin.reserve_resource, 'reservation_id', inputs)
        self.assertRaises(mgr_exceptions.MalformedParameter,
                          plugin.update_reservation, 'reservation_id', inputs)

    @ddt.data(-1, 0, '0', 'one')
    def test_error_with_amount(self, value):
        plugin = instance_plugin.VirtualInstancePlugin()
        inputs = self.get_input_values(2, 4018, 10, value, False,
                                       '2030-01-01 08:00', '2030-01-01 08:00',
                                       'lease-1', '')
        self.assertRaises(mgr_exceptions.MalformedParameter,
                          plugin.reserve_resource, 'reservation_id', inputs)
        self.assertRaises(mgr_exceptions.MalformedParameter,
                          plugin.update_reservation, 'reservation_id', inputs)

    @ddt.data('vcpus', 'memory_mb', 'disk_gb', 'amount', 'affinity',
              'resource_properties')
    def test_create_reservation_with_missing_param(self, missing_param):
        plugin = instance_plugin.VirtualInstancePlugin()
        inputs = self.get_input_values(2, 4018, 10, 1, False,
                                       '2030-01-01 08:00', '2030-01-01 08:00',
                                       'lease-1', '')
        del inputs[missing_param]
        self.assertRaises(mgr_exceptions.MissingParameter,
                          plugin.reserve_resource, 'reservation_id', inputs)

    def test_filter_hosts_by_reservation_with_exclude(self):
        def fake_get_reservation_by_host(host_id, start, end):
            if host_id == 'host-1':
                return [
                    {'id': '1',
                     'resource_type': instances.RESOURCE_TYPE}]
            else:
                return []

        hosts_list = [self.generate_host_info('host-1', 4, 4096, 1000),
                      self.generate_host_info('host-2', 4, 4096, 1000),
                      self.generate_host_info('host-3', 4, 4096, 1000)]
        mock_get_reservations = self.patch(db_utils,
                                           'get_reservations_by_host_id')

        mock_get_reservations.side_effect = fake_get_reservation_by_host
        free = [{'host': hosts_list[0], 'reservations': []},
                {'host': hosts_list[1], 'reservations': []},
                {'host': hosts_list[2], 'reservations': []}]
        non_free = []

        plugin = instance_plugin.VirtualInstancePlugin()
        ret = plugin.filter_hosts_by_reservation(hosts_list,
                                                 '2030-01-01 08:00',
                                                 '2030-01-01 12:00', ['1'])
        self.assertEqual(free, ret[0])
        self.assertEqual(non_free, ret[1])

    def test_pickup_host_from_reserved_hosts(self):
        def fake_max_usages(host, reservations):
            if host['id'] == 'host-1':
                return 4, 4096, 2000
            else:
                return 0, 0, 0

        def fake_get_reservation_by_host(host_id, start, end):
            return [
                {'id': '1', 'resource_type': instances.RESOURCE_TYPE},
                {'id': '2', 'resource_type': instances.RESOURCE_TYPE}]

        plugin = instance_plugin.VirtualInstancePlugin()

        mock_host_allocation_get = self.patch(
            db_api, 'host_allocation_get_all_by_values')
        mock_host_allocation_get.return_value = []

        mock_host_get_query = self.patch(db_api,
                                         'reservable_host_get_all_by_queries')
        hosts_list = [self.generate_host_info('host-1', 4, 4096, 1000),
                      self.generate_host_info('host-2', 4, 4096, 1000),
                      self.generate_host_info('host-3', 4, 4096, 1000)]
        mock_host_get_query.return_value = hosts_list

        mock_get_reservations = self.patch(db_utils,
                                           'get_reservations_by_host_id')

        mock_get_reservations.side_effect = fake_get_reservation_by_host
        plugin.max_usages = fake_max_usages

        mock_reservation_get = self.patch(db_api, 'reservation_get')
        mock_reservation_get.return_value = {
            'status': 'pending'
            }

        values = {
            'vcpus': 1,
            'memory_mb': 1024,
            'disk_gb': 20,
            'amount': 2,
            'affinity': False,
            'resource_properties': '',
            'start_date': datetime.datetime(2030, 1, 1, 8, 00),
            'end_date': datetime.datetime(2030, 1, 1, 12, 00)
            }
        expected = {'added': ['host-2', 'host-3'], 'removed': []}
        ret = plugin.pickup_hosts('reservation-id1', values)

        self.assertEqual(expected, ret)
        expected_query = ['vcpus >= 1', 'memory_mb >= 1024', 'local_gb >= 20']
        mock_host_get_query.assert_called_once_with(expected_query)

    def test_pickup_host_from_free_hosts(self):
        def fake_get_reservation_by_host(host_id, start, end):
            return []

        plugin = instance_plugin.VirtualInstancePlugin()

        mock_host_allocation_get = self.patch(
            db_api, 'host_allocation_get_all_by_values')
        mock_host_allocation_get.return_value = []

        mock_host_get_query = self.patch(db_api,
                                         'reservable_host_get_all_by_queries')
        hosts_list = [self.generate_host_info('host-1', 4, 4096, 1000),
                      self.generate_host_info('host-2', 4, 4096, 1000),
                      self.generate_host_info('host-3', 4, 4096, 1000)]
        mock_host_get_query.return_value = hosts_list

        mock_get_reservations = self.patch(db_utils,
                                           'get_reservations_by_host_id')
        mock_get_reservations.side_effect = fake_get_reservation_by_host

        mock_reservation_get = self.patch(db_api, 'reservation_get')
        mock_reservation_get.return_value = {
            'status': 'pending'
            }

        values = {
            'vcpus': 1,
            'memory_mb': 1024,
            'disk_gb': 20,
            'amount': 2,
            'affinity': False,
            'resource_properties': '',
            'start_date': datetime.datetime(2030, 1, 1, 8, 00),
            'end_date': datetime.datetime(2030, 1, 1, 12, 00),
            }
        expected = {'added': ['host-1', 'host-2'], 'removed': []}
        ret = plugin.pickup_hosts('reservation-id1', values)

        self.assertEqual(expected, ret)
        expected_query = ['vcpus >= 1', 'memory_mb >= 1024', 'local_gb >= 20']
        mock_host_get_query.assert_called_once_with(expected_query)

    def test_pickup_host_from_free_and_reserved_host(self):
        def fake_get_reservation_by_host(host_id, start, end):
            if host_id in ['host-1', 'host-3']:
                return [
                    {'id': '1',
                     'resource_type': instances.RESOURCE_TYPE},
                    {'id': '2',
                     'resource_type': instances.RESOURCE_TYPE}
                    ]
            else:
                return []

        plugin = instance_plugin.VirtualInstancePlugin()

        mock_host_allocation_get = self.patch(
            db_api, 'host_allocation_get_all_by_values')
        mock_host_allocation_get.return_value = []

        mock_host_get_query = self.patch(db_api,
                                         'reservable_host_get_all_by_queries')
        hosts_list = [self.generate_host_info('host-1', 4, 4096, 1000),
                      self.generate_host_info('host-2', 4, 4096, 1000),
                      self.generate_host_info('host-3', 4, 4096, 1000)]
        mock_host_get_query.return_value = hosts_list

        mock_get_reservations = self.patch(db_utils,
                                           'get_reservations_by_host_id')

        mock_get_reservations.side_effect = fake_get_reservation_by_host

        mock_max_usages = self.patch(plugin, 'max_usages')
        mock_max_usages.return_value = (0, 0, 0)

        mock_reservation_get = self.patch(db_api, 'reservation_get')
        mock_reservation_get.return_value = {
            'status': 'pending'
            }

        params = {
            'vcpus': 1,
            'memory_mb': 1024,
            'disk_gb': 20,
            'amount': 2,
            'affinity': False,
            'resource_properties': '',
            'start_date': datetime.datetime(2030, 1, 1, 8, 00),
            'end_date': datetime.datetime(2030, 1, 1, 12, 00)
            }

        expected = {'added': ['host-1', 'host-3'], 'removed': []}
        ret = plugin.pickup_hosts('reservation-id1', params)

        self.assertEqual(expected, ret)
        expected_query = ['vcpus >= 1', 'memory_mb >= 1024', 'local_gb >= 20']
        mock_host_get_query.assert_called_once_with(expected_query)

    def test_pickup_host_with_affinity(self):
        def fake_get_reservation_by_host(host_id, start, end):
            if host_id in ['host-1', 'host-3']:
                return [
                    {'id': '1',
                     'resource_type': instances.RESOURCE_TYPE},
                    {'id': '2',
                     'resource_type': instances.RESOURCE_TYPE}
                    ]
            else:
                return []

        plugin = instance_plugin.VirtualInstancePlugin()

        mock_host_allocation_get = self.patch(
            db_api, 'host_allocation_get_all_by_values')
        mock_host_allocation_get.return_value = []

        mock_host_get_query = self.patch(db_api,
                                         'reservable_host_get_all_by_queries')
        hosts_list = [self.generate_host_info('host-1', 8, 8192, 1000),
                      self.generate_host_info('host-2', 2, 2048, 500),
                      self.generate_host_info('host-3', 2, 2048, 500)]
        mock_host_get_query.return_value = hosts_list

        mock_get_reservations = self.patch(db_utils,
                                           'get_reservations_by_host_id')

        mock_get_reservations.side_effect = fake_get_reservation_by_host

        mock_max_usages = self.patch(plugin, 'max_usages')
        mock_max_usages.return_value = (0, 0, 0)

        mock_reservation_get = self.patch(db_api, 'reservation_get')
        mock_reservation_get.return_value = {
            'status': 'pending'
            }

        params = {
            'vcpus': 2,
            'memory_mb': 2048,
            'disk_gb': 100,
            'amount': 2,
            'affinity': True,
            'resource_properties': '',
            'start_date': datetime.datetime(2030, 1, 1, 8, 00),
            'end_date': datetime.datetime(2030, 1, 1, 12, 00)
            }

        expected = {'added': ['host-1', 'host-1'], 'removed': []}
        ret = plugin.pickup_hosts('reservation-id1', params)

        self.assertEqual(expected, ret)
        expected_query = ['vcpus >= 2', 'memory_mb >= 2048', 'local_gb >= 100']
        mock_host_get_query.assert_called_once_with(expected_query)

    def test_pickup_host_with_anti_affinity(self):
        def fake_get_reservation_by_host(host_id, start, end):
            if host_id in ['host-1', 'host-3']:
                return [
                    {'id': '1',
                     'resource_type': instances.RESOURCE_TYPE},
                    {'id': '2',
                     'resource_type': instances.RESOURCE_TYPE}
                    ]
            else:
                return []

        plugin = instance_plugin.VirtualInstancePlugin()

        mock_host_allocation_get = self.patch(
            db_api, 'host_allocation_get_all_by_values')
        mock_host_allocation_get.return_value = []

        mock_host_get_query = self.patch(db_api,
                                         'reservable_host_get_all_by_queries')
        hosts_list = [self.generate_host_info('host-1', 8, 8192, 1000),
                      self.generate_host_info('host-2', 2, 2048, 500)]
        mock_host_get_query.return_value = hosts_list

        mock_get_reservations = self.patch(db_utils,
                                           'get_reservations_by_host_id')

        mock_get_reservations.side_effect = fake_get_reservation_by_host

        mock_max_usages = self.patch(plugin, 'max_usages')
        mock_max_usages.return_value = (0, 0, 0)

        mock_reservation_get = self.patch(db_api, 'reservation_get')
        mock_reservation_get.return_value = {
            'status': 'pending'
            }

        params = {
            'vcpus': 2,
            'memory_mb': 2048,
            'disk_gb': 100,
            'amount': 2,
            'affinity': False,
            'resource_properties': '',
            'start_date': datetime.datetime(2030, 1, 1, 8, 00),
            'end_date': datetime.datetime(2030, 1, 1, 12, 00)
            }

        expected = {'added': ['host-1', 'host-2'], 'removed': []}
        ret = plugin.pickup_hosts('reservation-id1', params)

        self.assertEqual(expected, ret)
        expected_query = ['vcpus >= 2', 'memory_mb >= 2048', 'local_gb >= 100']
        mock_host_get_query.assert_called_once_with(expected_query)

    @ddt.data('None', 'none', None)
    def test_pickup_host_with_no_affinity(self, value):
        def fake_get_reservation_by_host(host_id, start, end):
            return []

        plugin = instance_plugin.VirtualInstancePlugin()

        mock_host_allocation_get = self.patch(
            db_api, 'host_allocation_get_all_by_values')
        mock_host_allocation_get.return_value = []

        mock_host_get_query = self.patch(db_api,
                                         'reservable_host_get_all_by_queries')
        hosts_list = [self.generate_host_info('host-1', 8, 8192, 1000),
                      self.generate_host_info('host-2', 2, 2048, 500),
                      self.generate_host_info('host-3', 2, 2048, 500)]
        mock_host_get_query.return_value = hosts_list

        mock_get_reservations = self.patch(db_utils,
                                           'get_reservations_by_host_id')

        mock_get_reservations.side_effect = fake_get_reservation_by_host

        mock_max_usages = self.patch(plugin, 'max_usages')
        mock_max_usages.return_value = (0, 0, 0)

        mock_reservation_get = self.patch(db_api, 'reservation_get')
        mock_reservation_get.return_value = {
            'status': 'pending'
            }

        params = {
            'vcpus': 4,
            'memory_mb': 4096,
            'disk_gb': 200,
            'amount': 2,
            'affinity': value,
            'resource_properties': '',
            'start_date': datetime.datetime(2030, 1, 1, 8, 00),
            'end_date': datetime.datetime(2030, 1, 1, 12, 00)
            }

        expected = {'added': ['host-1', 'host-1'], 'removed': []}
        ret = plugin.pickup_hosts('reservation-id1', params)

        self.assertEqual(expected, ret)
        expected_query = ['vcpus >= 4', 'memory_mb >= 4096', 'local_gb >= 200']
        mock_host_get_query.assert_called_once_with(expected_query)

    def test_pickup_host_from_less_hosts(self):
        def fake_get_reservation_by_host(host_id, start, end):
            if host_id in ['host-1', 'host-3']:
                return [
                    {'id': '1',
                     'resource_type': oshosts.RESOURCE_TYPE},
                    {'id': '2',
                     'resource_type': instances.RESOURCE_TYPE}
                    ]
            else:
                return [
                    {'id': '1',
                     'resource_type': instances.RESOURCE_TYPE},
                    {'id': '2',
                     'resource_type': instances.RESOURCE_TYPE}
                    ]

        plugin = instance_plugin.VirtualInstancePlugin()

        mock_host_get_query = self.patch(db_api,
                                         'reservable_host_get_all_by_queries')
        hosts_list = [self.generate_host_info('host-1', 4, 4096, 1000),
                      self.generate_host_info('host-2', 4, 4096, 1000),
                      self.generate_host_info('host-3', 4, 4096, 1000)]
        mock_host_get_query.return_value = hosts_list

        mock_get_reservations = self.patch(db_utils,
                                           'get_reservations_by_host_id')

        mock_get_reservations.side_effect = fake_get_reservation_by_host
        mock_host_allocation_get = self.patch(
            db_api, 'host_allocation_get_all_by_values')
        mock_host_allocation_get.return_value = []

        old_reservation = {
            'id': 'reservation-id1',
            'status': 'pending',
            'lease_id': 'lease-id1',
            'resource_id': 'instance-reservation-id1',
            'vcpus': 2, 'memory_mb': 1024, 'disk_gb': 100,
            'amount': 2, 'affinity': False,
            'resource_properties': ''}
        mock_reservation_get = self.patch(db_api, 'reservation_get')
        mock_reservation_get.return_value = old_reservation

        mock_lease_get = self.patch(db_api, 'lease_get')
        mock_lease_get.return_value = {'start_date': '2030-01-01 8:00',
                                       'end_date': '2030-01-01 12:00'}

        mock_max_usages = self.patch(plugin, 'max_usages')
        mock_max_usages.return_value = (1, 1024, 100)

        values = {
            'vcpus': 1,
            'memory_mb': 1024,
            'disk_gb': 20,
            'amount': 2,
            'affinity': False,
            'resource_properties': '',
            'start_date': datetime.datetime(2030, 1, 1, 8, 00),
            'end_date': datetime.datetime(2030, 1, 1, 12, 00)
            }

        self.assertRaises(mgr_exceptions.NotEnoughHostsAvailable,
                          plugin.update_reservation, 'reservation-id1',
                          values)

    def test_max_usage_with_serial_reservation(self):
        def fake_event_get(sort_key, sort_dir, filters):
            if filters['lease_id'] == 'lease-1':
                return self.generate_basic_events('lease-1',
                                                  '2030-01-01 08:00',
                                                  '2030-01-01 10:00',
                                                  '2030-01-01 11:00')
            elif filters['lease_id'] == 'lease-2':
                return self.generate_basic_events('lease-2',
                                                  '2030-01-01 12:00',
                                                  '2030-01-01 13:00',
                                                  '2030-01-01 14:00')

        plugin = instance_plugin.VirtualInstancePlugin()
        reservations = [
            {
                'lease_id': 'lease-1',
                'instance_reservation': {
                    'vcpus': 2, 'memory_mb': 3072, 'disk_gb': 20}},
            {
                'lease_id': 'lease-2',
                'instance_reservation': {
                    'vcpus': 3, 'memory_mb': 2048, 'disk_gb': 30}}
            ]

        mock_event_get = self.patch(db_api, 'event_get_all_sorted_by_filters')
        mock_event_get.side_effect = fake_event_get

        expected = (3, 3072, 30)
        ret = plugin.max_usages('fake-host', reservations)

        self.assertEqual(expected, ret)

    def test_max_usage_with_parallel_reservation(self):
        def fake_event_get(sort_key, sort_dir, filters):
            if filters['lease_id'] == 'lease-1':
                return self.generate_basic_events('lease-1',
                                                  '2030-01-01 08:00',
                                                  '2030-01-01 10:00',
                                                  '2030-01-01 11:00')
            elif filters['lease_id'] == 'lease-2':
                return self.generate_basic_events('lease-2',
                                                  '2030-01-01 10:00',
                                                  '2030-01-01 13:00',
                                                  '2030-01-01 14:00')

        plugin = instance_plugin.VirtualInstancePlugin()
        reservations = [
            {
                'lease_id': 'lease-1',
                'instance_reservation': {
                    'vcpus': 2, 'memory_mb': 3072, 'disk_gb': 20}},
            {
                'lease_id': 'lease-2',
                'instance_reservation': {
                    'vcpus': 3, 'memory_mb': 2048, 'disk_gb': 30}},
            ]

        mock_event_get = self.patch(db_api, 'event_get_all_sorted_by_filters')
        mock_event_get.side_effect = fake_event_get

        expected = (5, 5120, 50)
        ret = plugin.max_usages('fake-host', reservations)

        self.assertEqual(expected, ret)

    def test_max_usage_with_multi_reservation(self):
        def fake_event_get(sort_key, sort_dir, filters):
            if filters['lease_id'] == 'lease-1':
                return self.generate_basic_events('lease-1',
                                                  '2030-01-01 08:00',
                                                  '2030-01-01 10:00',
                                                  '2030-01-01 11:00')

        plugin = instance_plugin.VirtualInstancePlugin()
        reservations = [
            {
                'lease_id': 'lease-1',
                'instance_reservation': {
                    'vcpus': 2, 'memory_mb': 3072, 'disk_gb': 20}},
            {
                'lease_id': 'lease-1',
                'instance_reservation': {
                    'vcpus': 3, 'memory_mb': 2048, 'disk_gb': 30}},
            ]

        mock_event_get = self.patch(db_api, 'event_get_all_sorted_by_filters')
        mock_event_get.side_effect = fake_event_get

        expected = (5, 5120, 50)
        ret = plugin.max_usages('fake-host', reservations)

        self.assertEqual(expected, ret)

    def test_max_usage_with_decrease_reservation(self):
        def fake_event_get(sort_key, sort_dir, filters):
            if filters['lease_id'] == 'lease-1':
                return self.generate_basic_events('lease-1',
                                                  '2030-01-01 08:00',
                                                  '2030-01-01 10:00',
                                                  '2030-01-01 11:00')
            elif filters['lease_id'] == 'lease-2':
                return self.generate_basic_events('lease-2',
                                                  '2030-01-01 10:00',
                                                  '2030-01-01 13:00',
                                                  '2030-01-01 14:00')
            elif filters['lease_id'] == 'lease-3':
                return self.generate_basic_events('lease-3',
                                                  '2030-01-01 15:00',
                                                  '2030-01-01 16:00',
                                                  '2030-01-01 17:00')

        plugin = instance_plugin.VirtualInstancePlugin()
        reservations = [
            {
                'lease_id': 'lease-1',
                'instance_reservation': {
                    'vcpus': 2, 'memory_mb': 3072, 'disk_gb': 20}},
            {
                'lease_id': 'lease-2',
                'instance_reservation': {
                    'vcpus': 1, 'memory_mb': 1024, 'disk_gb': 10}},
            {
                'lease_id': 'lease-3',
                'instance_reservation': {
                    'vcpus': 4, 'memory_mb': 2048, 'disk_gb': 40
                    }},
            ]

        mock_event_get = self.patch(db_api, 'event_get_all_sorted_by_filters')
        mock_event_get.side_effect = fake_event_get

        expected = (4, 4096, 40)
        ret = plugin.max_usages('fake-host', reservations)

        self.assertEqual(expected, ret)

    def test_create_resources(self):
        instance_reservation = {
            'reservation_id': 'reservation-id1',
            'vcpus': 2,
            'memory_mb': 1024,
            'disk_gb': 20,
            'affinity': False
            }

        plugin = instance_plugin.VirtualInstancePlugin()

        fake_client = mock.MagicMock()
        mock_nova_client = self.patch(nova, 'NovaClientWrapper')
        mock_nova_client.return_value = fake_client
        fake_server_group = mock.MagicMock(id='server_group_id1')
        fake_client.nova.server_groups.create.return_value = \
            fake_server_group

        self.set_context(context.BlazarContext(project_id='fake-project',
                                               auth_token='fake-token'))
        fake_flavor = mock.MagicMock(method='set_keys',
                                     flavorid='reservation-id1')
        mock_nova = mock.MagicMock()
        type(plugin).nova = mock_nova
        mock_nova.nova.flavors.create.return_value = fake_flavor

        mock_create_reservation_class = self.patch(
            plugin.placement_client, 'create_reservation_class')

        fake_pool = mock.MagicMock(id='pool-id1')
        fake_agg = mock.MagicMock()
        fake_pool.create.return_value = fake_agg
        mock_pool = self.patch(nova, 'ReservationPool')
        mock_pool.return_value = fake_pool

        expected = (fake_flavor, fake_server_group, fake_agg)

        ret = plugin._create_resources(instance_reservation)

        self.assertEqual(expected, ret)

        fake_client.nova.server_groups.create.assert_called_once_with(
            'reservation:reservation-id1', 'anti-affinity')
        mock_nova.nova.flavors.create.assert_called_once_with(
            flavorid='reservation-id1',
            name='reservation:reservation-id1',
            vcpus=2, ram=1024, disk=20, is_public=False)
        fake_flavor.set_keys.assert_called_once_with(
            {'aggregate_instance_extra_specs:reservation': 'reservation-id1',
             'affinity_id': 'server_group_id1',
             'resources:CUSTOM_RESERVATION_RESERVATION_ID1': '1'})
        fake_pool.create.assert_called_once_with(
            name='reservation-id1',
            metadata={'reservation': 'reservation-id1',
                      'filter_tenant_id': 'fake-project',
                      'affinity_id': 'server_group_id1'})
        mock_create_reservation_class.assert_called_once_with(
            'reservation-id1')

    def test_query_available_hosts(self):
        mock_host_get_query = self.patch(db_api,
                                         'reservable_host_get_all_by_queries')
        host1, host2, host3 = (self.generate_host_info(host_id, 4, 4096, 1000)
                               for host_id in ['host-1', 'host-2', 'host-3'])
        hosts_list = [host1, host2, host3]
        mock_host_get_query.return_value = hosts_list

        get_reservations = self.patch(db_utils,
                                      'get_reservations_by_host_id')
        get_reservations.return_value = []

        plugin = instance_plugin.VirtualInstancePlugin()

        query_params = {
            'cpus': 1, 'memory': 1024, 'disk': 10,
            'resource_properties': '',
            'start_date': datetime.datetime(2020, 7, 7, 18, 0),
            'end_date': datetime.datetime(2020, 7, 7, 19, 0)
        }

        ret = plugin.query_available_hosts(**query_params)

        expected = [host1] * 4 + [host2] * 4 + [host3] * 4
        self.assertEqual(expected, ret)

    def test_pickup_hosts_for_update(self):
        reservation = {'id': 'reservation-id1', 'status': 'pending'}
        plugin = instance_plugin.VirtualInstancePlugin()

        mock_alloc_get = self.patch(db_api,
                                    'host_allocation_get_all_by_values')
        mock_alloc_get.return_value = [
            {'compute_host_id': 'host-id1'}, {'compute_host_id': 'host-id2'},
            {'compute_host_id': 'host-id3'}]
        mock_query_available = self.patch(plugin, 'query_available_hosts')
        mock_query_available.return_value = [
            self.generate_host_info('host-id2', 2, 2024, 1000),
            self.generate_host_info('host-id3', 2, 2024, 1000),
            self.generate_host_info('host-id4', 2, 2024, 1000)]

        mock_reservation_get = self.patch(db_api, 'reservation_get')
        mock_reservation_get.return_value = reservation

        # case: new amount is less than old amount
        values = self.get_input_values(1, 1024, 10, 1, False,
                                       '2020-07-01 10:00', '2020-07-01 11:00',
                                       'lease-1', '')
        expect = {'added': [],
                  'removed': ['host-id1', 'host-id2', 'host-id3']}
        ret = plugin.pickup_hosts(reservation['id'], values)
        self.assertEqual(expect['added'], ret['added'])
        self.assertEqual(2, len(ret['removed']))
        self.assertTrue(all([h in expect['removed'] for h in ret['removed']]))
        query_params = {
            'cpus': 1, 'memory': 1024, 'disk': 10,
            'resource_properties': '',
            'start_date': '2020-07-01 10:00',
            'end_date': '2020-07-01 11:00',
            'excludes_res': ['reservation-id1']
            }
        mock_query_available.assert_called_with(**query_params)

        # case: new amount is same but change allocations
        values = self.get_input_values(1, 1024, 10, 3, False,
                                       '2020-07-01 10:00', '2020-07-01 11:00',
                                       'lease-1', '["==", "key1", "value1"]')
        expect = {'added': ['host-id4'], 'removed': ['host-id1']}
        ret = plugin.pickup_hosts(reservation['id'], values)
        self.assertEqual(expect['added'], ret['added'])
        self.assertEqual(expect['removed'], ret['removed'])
        query_params = {
            'cpus': 1, 'memory': 1024, 'disk': 10,
            'resource_properties': '["==", "key1", "value1"]',
            'start_date': '2020-07-01 10:00',
            'end_date': '2020-07-01 11:00',
            'excludes_res': ['reservation-id1']
            }
        mock_query_available.assert_called_with(**query_params)

        # case: new amount is greater than old amount
        host_ids = ('host-id1', 'host-id2', 'host-id3', 'host-id4')
        mock_query_available.return_value = [
            self.generate_host_info(host_id, 2, 2024, 1000)
            for host_id in host_ids]

        values = self.get_input_values(1, 1024, 10, 4, False,
                                       '2020-07-01 10:00', '2020-07-01 11:00',
                                       'lease-1', '')
        expect = {'added': ['host-id4'], 'removed': []}
        ret = plugin.pickup_hosts(reservation['id'], values)
        self.assertEqual(expect['added'], ret['added'])
        self.assertEqual(expect['removed'], ret['removed'])
        query_params = {
            'cpus': 1, 'memory': 1024, 'disk': 10,
            'resource_properties': '',
            'start_date': '2020-07-01 10:00',
            'end_date': '2020-07-01 11:00',
            'excludes_res': ['reservation-id1']
            }
        mock_query_available.assert_called_with(**query_params)

        # case: affinity is changed to True
        mock_query_available.return_value = [
            self.generate_host_info(host_id, 8, 8192, 1000)
            for host_id in host_ids * 8]

        values = self.get_input_values(1, 1024, 10, 4, True,
                                       '2020-07-01 10:00', '2020-07-01 11:00',
                                       'lease-1', '')
        ret = plugin.pickup_hosts(reservation['id'], values)

        # We don't care which host id (1-3) is picked up
        # Just make sure the same host is returned three times in "added"
        added = ret['added']
        self.assertEqual(3, len(added))
        self.assertEqual(1, len(set(added)))
        self.assertIn(added[0], ('host-id1', 'host-id2', 'host-id3'))

        # and make sure the other two hosts are removed
        removed = ret['removed']
        self.assertEqual(2, len(removed))
        self.assertEqual(2, len(set(removed)))
        expect_removed = set(host_ids) - set(added)
        for host_id in removed:
            self.assertIn(host_id, expect_removed)

        query_params = {
            'cpus': 1, 'memory': 1024, 'disk': 10,
            'resource_properties': '',
            'start_date': '2020-07-01 10:00',
            'end_date': '2020-07-01 11:00',
            'excludes_res': ['reservation-id1']
        }
        mock_query_available.assert_called_with(**query_params)

    def test_update_resources(self):
        reservation = {
            'id': 'reservation-id1',
            'status': 'pending',
            'vcpus': 2, 'memory_mb': 1024,
            'disk_gb': 10, 'server_group_id': 'group-1'}
        mock_reservation_get = self.patch(db_api, 'reservation_get')
        mock_reservation_get.return_value = reservation
        fake_client = mock.MagicMock()
        mock_nova_client = self.patch(nova, 'NovaClientWrapper')
        mock_nova_client.return_value = fake_client
        self.set_context(context.BlazarContext(project_id='fake-project',
                                               auth_token='fake-token'))
        plugin = instance_plugin.VirtualInstancePlugin()
        fake_flavor = mock.MagicMock(method='set_keys',
                                     flavorid='reservation-id1')
        mock_nova = mock.MagicMock()
        type(plugin).nova = mock_nova
        mock_nova.nova.flavors.create.return_value = fake_flavor

        plugin.update_resources('reservation-id1')

        mock_reservation_get.assert_called_once_with('reservation-id1')
        mock_nova.nova.flavors.delete.assert_called_once_with(
            'reservation-id1')
        mock_nova.nova.flavors.create.assert_called_once_with(
            flavorid='reservation-id1',
            name='reservation:reservation-id1',
            vcpus=2, ram=1024, disk=10, is_public=False)
        fake_flavor.set_keys.assert_called_once_with(
            {'aggregate_instance_extra_specs:reservation': 'reservation-id1',
             'affinity_id': 'group-1',
             'resources:CUSTOM_RESERVATION_RESERVATION_ID1': '1'})

    def test_update_resources_in_active(self):
        def fake_host_get(host_id):
            return {'service_name': 'host' + host_id[-1],
                    'hypervisor_hostname': 'hypvsr' + host_id[-1]}

        reservation = {
            'id': 'reservation-id1',
            'status': 'active',
            'vcpus': 2, 'memory_mb': 1024,
            'disk_gb': 10, 'aggregate_id': 'aggregate-1'}

        mock_reservation_get = self.patch(db_api, 'reservation_get')
        mock_reservation_get.return_value = reservation
        self.set_context(context.BlazarContext(project_id='fake-project'))
        plugin = instance_plugin.VirtualInstancePlugin()

        mock_update_reservation_inventory = self.patch(
            plugin.placement_client, 'update_reservation_inventory')

        fake_pool = mock.MagicMock()
        mock_pool = self.patch(nova, 'ReservationPool')
        mock_pool.return_value = fake_pool

        mock_alloc_get = self.patch(db_api,
                                    'host_allocation_get_all_by_values')
        mock_alloc_get.return_value = [
            {'compute_host_id': 'host-id1'}, {'compute_host_id': 'host-id2'},
            {'compute_host_id': 'host-id3'}, {'compute_host_id': 'host-id3'}]

        mock_host_get = self.patch(db_api, 'host_get')
        mock_host_get.side_effect = fake_host_get

        plugin.update_resources('reservation-id1')

        mock_reservation_get.assert_called_once_with('reservation-id1')
        for i in range(3):
            fake_pool.add_computehost.assert_any_call(
                'aggregate-1', 'host' + str(i + 1), stay_in=True)

        mock_update_reservation_inventory.assert_any_call(
            'hypvsr1', 'reservation-id1', 1)
        mock_update_reservation_inventory.assert_any_call(
            'hypvsr2', 'reservation-id1', 1)
        mock_update_reservation_inventory.assert_any_call(
            'hypvsr3', 'reservation-id1', 2)

    def test_update_reservation(self):
        plugin = instance_plugin.VirtualInstancePlugin()

        old_reservation = {
            'id': 'reservation-id1',
            'status': 'pending',
            'lease_id': 'lease-id1',
            'resource_id': 'instance-reservation-id1',
            'vcpus': 2, 'memory_mb': 1024, 'disk_gb': 100,
            'amount': 2, 'affinity': False,
            'resource_properties': ''}
        mock_reservation_get = self.patch(db_api, 'reservation_get')
        mock_reservation_get.return_value = old_reservation

        mock_lease_get = self.patch(db_api, 'lease_get')
        mock_lease_get.return_value = {'start_date': '2020-07-07 18:00',
                                       'end_date': '2020-07-07 19:00'}

        mock_pickup_hosts = self.patch(plugin, 'pickup_hosts')
        mock_pickup_hosts.return_value = {
            'added': set(['host-id1']), 'removed': set(['host-id2'])}

        mock_inst_update = self.patch(db_api, 'instance_reservation_update')
        mock_inst_update.return_value = {
            'vcpus': 4, 'memory_mb': 1024, 'disk_gb': 200,
            'amount': 2, 'affinity': False}

        mock_update_alloc = self.patch(plugin, 'update_host_allocations')

        mock_update_resource = self.patch(plugin, 'update_resources')

        new_values = {'vcpus': 4, 'disk_gb': 200}
        plugin.update_reservation('reservation-id1', new_values)

        mock_pickup_hosts.assert_called_once_with(
            'reservation-id1',
            {'vcpus': 4, 'memory_mb': 1024, 'disk_gb': 200,
             'amount': 2, 'affinity': False, 'resource_properties': ''})
        mock_inst_update.assert_called_once_with(
            'instance-reservation-id1',
            {'vcpus': 4, 'memory_mb': 1024, 'disk_gb': 200,
             'amount': 2, 'affinity': False, 'resource_properties': ''})
        mock_update_alloc.assert_called_once_with(set(['host-id1']),
                                                  set(['host-id2']),
                                                  'reservation-id1')
        mock_update_resource.assert_called_once_with('reservation-id1')

    def test_update_reservation_not_enough_hosts(self):
        plugin = instance_plugin.VirtualInstancePlugin()

        old_reservation = {
            'id': 'reservation-id1',
            'status': 'pending',
            'lease_id': 'lease-id1',
            'resource_id': 'instance-reservation-id1',
            'vcpus': 2, 'memory_mb': 1024, 'disk_gb': 100,
            'amount': 2, 'affinity': False,
            'resource_properties': ''}
        mock_reservation_get = self.patch(db_api, 'reservation_get')
        mock_reservation_get.return_value = old_reservation

        mock_lease_get = self.patch(db_api, 'lease_get')
        mock_lease_get.return_value = {'start_date': '2020-07-07 18:00',
                                       'end_date': '2020-07-07 19:00'}

        # Mock that we have at least two hosts for (2 vcpus + 100 disk_gb),
        # but we have only one for (4 vcpus and 200 disk_gb)
        mock_alloc_get = self.patch(db_api,
                                    'host_allocation_get_all_by_values')
        mock_alloc_get.return_value = [{'compute_host_id': 'host-id1'},
                                       {'compute_host_id': 'host-id2'}]

        mock_query_available = self.patch(plugin, 'query_available_hosts')
        mock_query_available.return_value = [
            self.generate_host_info('host-id1', 4, 2048, 1000)]

        new_values = {'vcpus': 4, 'disk_gb': 200,
                      'start_date': datetime.datetime(2020, 7, 7, 18, 0),
                      'end_date': datetime.datetime(2020, 7, 7, 19, 0),
                      'id': '00ee4f12-77c8-44d5-abca-06a543210a50'}
        self.assertRaises(mgr_exceptions.NotEnoughHostsAvailable,
                          plugin.update_reservation, 'reservation-id1',
                          new_values)

    def test_update_flavor_in_active(self):
        plugin = instance_plugin.VirtualInstancePlugin()

        old_reservation = {
            'id': 'reservation-id1',
            'status': 'active',
            'lease_id': 'lease-id1',
            'resource_id': 'instance-reservation-id1',
            'vcpus': 2, 'memory_mb': 1024, 'disk_gb': 100,
            'amount': 2, 'affinity': False}
        mock_reservation_get = self.patch(db_api, 'reservation_get')
        mock_reservation_get.return_value = old_reservation

        mock_lease_get = self.patch(db_api, 'lease_get')
        mock_lease_get.return_value = {'start_date': '2020-07-07 18:00',
                                       'end_date': '2020-07-07 19:00'}

        new_values = {'vcpus': 4, 'disk_gb': 200}
        self.assertRaises(mgr_exceptions.InvalidStateUpdate,
                          plugin.update_reservation,
                          'reservation-id1', new_values)

    def test_update_host_allocations(self):
        mock_alloc_get = self.patch(db_api,
                                    'host_allocation_get_all_by_values')
        mock_alloc_get.return_value = [
            {'id': 'id10', 'compute_host_id': 'host-id10'},
            {'id': 'id11', 'compute_host_id': 'host-id11'},
            {'id': 'id12', 'compute_host_id': 'host-id11'},
            {'id': 'id13', 'compute_host_id': 'host-id11'},
            {'id': 'id14', 'compute_host_id': 'host-id12'}]

        mock_alloc_destroy = self.patch(db_api, 'host_allocation_destroy')
        mock_alloc_create = self.patch(db_api, 'host_allocation_create')

        plugin = instance_plugin.VirtualInstancePlugin()

        added_host = ['host-id1', 'host-id1', 'host-id2']
        removed_host = ['host-id10', 'host-id11', 'host-id11']

        plugin.update_host_allocations(added_host, removed_host,
                                       'reservation-id1')

        removed_calls = [mock.call('id10'), mock.call('id11')]
        mock_alloc_destroy.assert_has_calls(removed_calls)
        self.assertEqual(3, mock_alloc_destroy.call_count)

        added_calls = [
            mock.call({'compute_host_id': 'host-id1',
                       'reservation_id': 'reservation-id1'}),
            mock.call({'compute_host_id': 'host-id2',
                       'reservation_id': 'reservation-id1'})]
        mock_alloc_create.assert_has_calls(added_calls)
        self.assertEqual(3, mock_alloc_create.call_count)

    def test_on_start(self):
        def fake_host_get(host_id):
            return {'service_name': 'host' + host_id[-1],
                    'hypervisor_hostname': 'hypvsr' + host_id[-1]}

        self.set_context(context.BlazarContext(project_id='fake-project'))
        plugin = instance_plugin.VirtualInstancePlugin()

        mock_inst_get = self.patch(db_api, 'instance_reservation_get')
        mock_inst_get.return_value = {'reservation_id': 'reservation-id1',
                                      'aggregate_id': 'aggregate-id1'}

        mock_nova = mock.MagicMock()
        type(plugin).nova = mock_nova

        fake_pool = mock.MagicMock()
        mock_pool = self.patch(nova, 'ReservationPool')
        mock_pool.return_value = fake_pool

        mock_update_reservation_inventory = self.patch(
            plugin.placement_client, 'update_reservation_inventory')

        mock_alloc_get = self.patch(db_api,
                                    'host_allocation_get_all_by_values')
        mock_alloc_get.return_value = [
            {'compute_host_id': 'host-id1'}, {'compute_host_id': 'host-id2'},
            {'compute_host_id': 'host-id3'}, {'compute_host_id': 'host-id3'}]

        mock_host_get = self.patch(db_api, 'host_get')
        mock_host_get.side_effect = fake_host_get

        plugin.on_start('resource-id1')

        mock_nova.flavor_access.add_tenant_access.assert_called_once_with(
            'reservation-id1', 'fake-project')
        for i in range(3):
            fake_pool.add_computehost.assert_any_call(
                'aggregate-id1', 'host' + str(i + 1), stay_in=True)

        mock_update_reservation_inventory.assert_any_call(
            'hypvsr1', 'reservation-id1', 1)
        mock_update_reservation_inventory.assert_any_call(
            'hypvsr2', 'reservation-id1', 1)
        mock_update_reservation_inventory.assert_any_call(
            'hypvsr3', 'reservation-id1', 2)

    def test_on_end(self):
        self.set_context(context.BlazarContext(project_id='fake-project-id'))

        plugin = instance_plugin.VirtualInstancePlugin()

        fake_instance_reservation = {'reservation_id': 'reservation-id1'}
        mock_inst_get = self.patch(db_api, 'instance_reservation_get')
        mock_inst_get.return_value = fake_instance_reservation

        mock_alloc_get = self.patch(db_api,
                                    'host_allocation_get_all_by_values')
        mock_alloc_get.return_value = [{'id': 'host-alloc-id1',
                                        'compute_host_id': 'host-id1'},
                                       {'id': 'host-alloc-id2',
                                        'compute_host_id': 'host-id2'}]

        mock_host_get = self.patch(db_api, 'host_get')
        mock_host_get.side_effect = [
            {'service_name': 'host1', 'hypervisor_hostname': 'hypvsr1'},
            {'service_name': 'host2', 'hypervisor_hostname': 'hypvsr2'}
        ]

        mock_delete_reservation_inventory = self.patch(
            plugin.placement_client, 'delete_reservation_inventory')
        mock_delete_reservation_class = self.patch(
            plugin.placement_client, 'delete_reservation_class')

        self.patch(db_api, 'host_allocation_destroy')

        fake_servers = [mock.MagicMock() for i in range(5)]
        mock_nova = mock.MagicMock()
        type(plugin).nova = mock_nova
        # First, we return the fake servers to delete. Second, on the check in
        # _check_server_deletion(), we mock they are still in nova DB to
        # exercise retry and at last we mock they are deleted completely.
        mock_nova.servers.list.side_effect = [fake_servers, fake_servers, []]

        mock_cleanup_resources = self.patch(plugin, 'cleanup_resources')

        mock_log = self.patch(instance_plugin, 'LOG')
        mock_nova.servers.delete.side_effect = [nova_exceptions.NotFound(
            404, "The server doesn't exist in Nova"), Exception('Unknown'),
            None, None, None]

        plugin.on_end('resource-id1')

        mock_nova.flavor_access.remove_tenant_access.assert_called_once_with(
            'reservation-id1', 'fake-project-id')

        mock_nova.servers.list.assert_called_with(
            search_opts={'flavor': 'reservation-id1', 'all_tenants': 1},
            detailed=False)
        mock_nova.servers.list.call_count = 3
        self.assertEqual(5, mock_nova.servers.delete.call_count)
        mock_log.info.assert_any_call(
            "Could not find server '%s', may have been deleted concurrently.",
            fake_servers[0].id)
        mock_log.exception.assert_called_with(
            "Failed to delete server '%s': %s.", fake_servers[1].id, 'Unknown')
        for i in range(2):
            mock_delete_reservation_inventory.assert_any_call(
                'hypvsr' + str(i + 1), 'reservation-id1')
        mock_cleanup_resources.assert_called_once_with(
            fake_instance_reservation)
        mock_delete_reservation_class.assert_called_once_with(
            'reservation-id1')

    def test_heal_reservations_before_start_and_resources_changed(self):
        plugin = instance_plugin.VirtualInstancePlugin()
        failed_host = {'id': '1'}
        dummy_reservation = {
            'id': 'rsrv-1',
            'resource_type': instances.RESOURCE_TYPE,
            'lease_id': 'lease-1',
            'status': 'pending',
            'vcpus': 2,
            'memory_mb': 1024,
            'disk_gb': 256,
            'aggregate_id': 'agg-1',
            'affinity': False,
            'amount': 3,
            'resource_properties': '',
            'computehost_allocations': [{
                'id': 'alloc-1', 'compute_host_id': failed_host['id'],
                'reservation_id': 'rsrv-1'
            }]
        }
        get_reservations = self.patch(db_utils,
                                      'get_reservations_by_host_ids')
        get_reservations.return_value = [dummy_reservation]
        heal_reservation = self.patch(plugin, '_heal_reservation')
        heal_reservation.return_value = True

        result = plugin.heal_reservations(
            [failed_host],
            datetime.datetime(2020, 1, 1, 12, 00),
            datetime.datetime(2020, 1, 1, 13, 00))
        heal_reservation.assert_called_once_with(
            dummy_reservation, list(failed_host.values()))
        self.assertEqual({}, result)

    def test_heal_reservations_before_start_and_missing_resources(self):
        plugin = instance_plugin.VirtualInstancePlugin()
        failed_host = {'id': '1'}
        dummy_reservation = {
            'id': 'rsrv-1',
            'resource_type': instances.RESOURCE_TYPE,
            'lease_id': 'lease-1',
            'status': 'pending',
            'vcpus': 2,
            'memory_mb': 1024,
            'disk_gb': 256,
            'aggregate_id': 'agg-1',
            'affinity': False,
            'amount': 3,
            'resource_properties': '',
            'computehost_allocations': [{
                'id': 'alloc-1', 'compute_host_id': failed_host['id'],
                'reservation_id': 'rsrv-1'
            }]
        }
        get_reservations = self.patch(db_utils,
                                      'get_reservations_by_host_ids')
        get_reservations.return_value = [dummy_reservation]
        heal_reservation = self.patch(plugin, '_heal_reservation')
        heal_reservation.return_value = False

        result = plugin.heal_reservations(
            [failed_host],
            datetime.datetime(2020, 1, 1, 12, 00),
            datetime.datetime(2020, 1, 1, 13, 00))
        heal_reservation.assert_called_once_with(
            dummy_reservation, list(failed_host.values()))
        self.assertEqual(
            {dummy_reservation['id']: {'missing_resources': True}},
            result)

    def test_heal_active_reservations_and_resources_changed(self):
        plugin = instance_plugin.VirtualInstancePlugin()
        failed_host = {'id': '1'}
        dummy_reservation = {
            'id': 'rsrv-1',
            'resource_type': instances.RESOURCE_TYPE,
            'lease_id': 'lease-1',
            'status': 'active',
            'vcpus': 2,
            'memory_mb': 1024,
            'disk_gb': 256,
            'aggregate_id': 'agg-1',
            'affinity': False,
            'amount': 3,
            'computehost_allocations': [{
                'id': 'alloc-1', 'compute_host_id': failed_host['id'],
                'reservation_id': 'rsrv-1'
            }]
        }
        get_reservations = self.patch(db_utils,
                                      'get_reservations_by_host_ids')
        get_reservations.return_value = [dummy_reservation]
        heal_reservation = self.patch(plugin, '_heal_reservation')
        heal_reservation.return_value = True

        result = plugin.heal_reservations(
            [failed_host],
            datetime.datetime(2020, 1, 1, 12, 00),
            datetime.datetime(2020, 1, 1, 13, 00))
        heal_reservation.assert_called_once_with(
            dummy_reservation, list(failed_host.values()))
        self.assertEqual(
            {dummy_reservation['id']: {'resources_changed': True}},
            result)

    def test_heal_active_reservations_and_missing_resources(self):
        plugin = instance_plugin.VirtualInstancePlugin()
        failed_host = {'id': '1'}
        dummy_reservation = {
            'id': 'rsrv-1',
            'resource_type': instances.RESOURCE_TYPE,
            'lease_id': 'lease-1',
            'status': 'active',
            'vcpus': 2,
            'memory_mb': 1024,
            'disk_gb': 256,
            'aggregate_id': 'agg-1',
            'affinity': False,
            'amount': 3,
            'computehost_allocations': [{
                'id': 'alloc-1', 'compute_host_id': failed_host['id'],
                'reservation_id': 'rsrv-1'
            }]
        }
        get_reservations = self.patch(db_utils,
                                      'get_reservations_by_host_ids')
        get_reservations.return_value = [dummy_reservation]
        heal_reservation = self.patch(plugin, '_heal_reservation')
        heal_reservation.return_value = False

        result = plugin.heal_reservations(
            [failed_host],
            datetime.datetime(2020, 1, 1, 12, 00),
            datetime.datetime(2020, 1, 1, 13, 00))
        heal_reservation.assert_called_once_with(
            dummy_reservation, list(failed_host.values()))
        self.assertEqual(
            {dummy_reservation['id']: {'missing_resources': True}},
            result)

    def test_reallocate_before_start(self):
        plugin = instance_plugin.VirtualInstancePlugin()
        failed_host = {'id': '1'}
        new_host = {'id': '2'}
        dummy_reservation = {
            'id': 'rsrv-1',
            'resource_type': instances.RESOURCE_TYPE,
            'lease_id': 'lease-1',
            'status': 'pending',
            'vcpus': 2,
            'memory_mb': 1024,
            'disk_gb': 256,
            'aggregate_id': 'agg-1',
            'affinity': False,
            'amount': 3,
            'resource_properties': '',
            'computehost_allocations': [{
                'id': 'alloc-1', 'compute_host_id': failed_host['id'],
                'reservation_id': 'rsrv-1'}]
        }
        dummy_lease = {
            'name': 'lease-name',
            'start_date': datetime.datetime(2020, 1, 1, 12, 00),
            'end_date': datetime.datetime(2020, 1, 2, 12, 00),
            'trust_id': 'trust-1'
        }
        lease_get = self.patch(db_api, 'lease_get')
        lease_get.return_value = dummy_lease
        pickup_hosts = self.patch(plugin, 'pickup_hosts')
        pickup_hosts.return_value = {'added': [new_host['id']], 'removed': []}
        alloc_update = self.patch(db_api, 'host_allocation_update')

        with mock.patch.object(datetime, 'datetime',
                               mock.Mock(wraps=datetime.datetime)) as patched:
            patched.utcnow.return_value = datetime.datetime(
                2020, 1, 1, 11, 00)
            result = plugin._heal_reservation(
                dummy_reservation, list(failed_host.values()))

        pickup_hosts.assert_called_once()
        alloc_update.assert_called_once_with(
            'alloc-1', {'compute_host_id': new_host['id']})
        self.assertEqual(True, result)

    def test_reallocate_active(self):
        plugin = instance_plugin.VirtualInstancePlugin()
        failed_host = {'id': '1',
                       'service_name': 'compute-1',
                       'hypervisor_hostname': 'hypvsr-1'}
        new_host = {'id': '2',
                    'service_name': 'compute-2',
                    'hypervisor_hostname': 'hypvsr-2'}
        dummy_reservation = {
            'id': 'rsrv-1',
            'resource_type': instances.RESOURCE_TYPE,
            'lease_id': 'lease-1',
            'status': 'active',
            'vcpus': 2,
            'memory_mb': 1024,
            'disk_gb': 256,
            'aggregate_id': 'agg-1',
            'affinity': False,
            'amount': 3,
            'resource_properties': '',
            'computehost_allocations': [{
                'id': 'alloc-1', 'compute_host_id': failed_host['id'],
                'reservation_id': 'rsrv-1'}]
        }
        dummy_lease = {
            'name': 'lease-name',
            'start_date': datetime.datetime(2020, 1, 1, 12, 00),
            'end_date': datetime.datetime(2020, 1, 2, 12, 00),
            'trust_id': 'trust-1'
        }
        reservation_get = self.patch(db_api, 'reservation_get')
        reservation_get.return_value = dummy_reservation
        lease_get = self.patch(db_api, 'lease_get')
        lease_get.return_value = dummy_lease
        host_get = self.patch(db_api, 'host_get')
        host_get.side_effect = [failed_host, new_host]
        fake_pool = mock.MagicMock()
        mock_pool = self.patch(nova, 'ReservationPool')
        mock_pool.return_value = fake_pool
        pickup_hosts = self.patch(plugin, 'pickup_hosts')
        pickup_hosts.return_value = {'added': [new_host['id']], 'removed': []}
        alloc_update = self.patch(db_api, 'host_allocation_update')
        mock_delete_reservation_inventory = self.patch(
            plugin.placement_client, 'delete_reservation_inventory')
        mock_update_reservation_inventory = self.patch(
            plugin.placement_client, 'update_reservation_inventory')

        with mock.patch.object(datetime, 'datetime',
                               mock.Mock(wraps=datetime.datetime)) as patched:
            patched.utcnow.return_value = datetime.datetime(
                2020, 1, 1, 13, 00)
            result = plugin._heal_reservation(
                dummy_reservation, list(failed_host.values()))

        fake_pool.remove_computehost.assert_called_once_with(
            dummy_reservation['aggregate_id'],
            failed_host['service_name'])
        pickup_hosts.assert_called_once()
        alloc_update.assert_called_once_with(
            'alloc-1', {'compute_host_id': new_host['id']})
        fake_pool.add_computehost.assert_called_once_with(
            dummy_reservation['aggregate_id'],
            new_host['service_name'],
            stay_in=True)
        mock_delete_reservation_inventory.assert_called_once_with(
            'hypvsr-1', 'rsrv-1')
        mock_update_reservation_inventory.assert_called_once_with(
            'hypvsr-2', 'rsrv-1', 1, additional=True)
        self.assertEqual(True, result)

    def test_reallocate_missing_resources(self):
        plugin = instance_plugin.VirtualInstancePlugin()
        failed_host = {'id': '1',
                       'service_name': 'compute-1'}
        dummy_reservation = {
            'id': 'rsrv-1',
            'resource_type': instances.RESOURCE_TYPE,
            'lease_id': 'lease-1',
            'status': 'pending',
            'vcpus': 2,
            'memory_mb': 1024,
            'disk_gb': 256,
            'aggregate_id': 'agg-1',
            'affinity': False,
            'amount': 3,
            'resource_properties': '',
            'computehost_allocations': [{
                'id': 'alloc-1', 'compute_host_id': failed_host['id'],
                'reservation_id': 'rsrv-1'}]
        }
        dummy_lease = {
            'name': 'lease-name',
            'start_date': datetime.datetime(2020, 1, 1, 12, 00),
            'end_date': datetime.datetime(2020, 1, 2, 12, 00),
            'trust_id': 'trust-1'
        }
        reservation_get = self.patch(db_api, 'reservation_get')
        reservation_get.return_value = dummy_reservation
        lease_get = self.patch(db_api, 'lease_get')
        lease_get.return_value = dummy_lease
        pickup_hosts = self.patch(plugin, 'pickup_hosts')
        pickup_hosts.side_effect = mgr_exceptions.NotEnoughHostsAvailable
        alloc_destroy = self.patch(db_api, 'host_allocation_destroy')

        with mock.patch.object(datetime, 'datetime',
                               mock.Mock(wraps=datetime.datetime)) as patched:
            patched.utcnow.return_value = datetime.datetime(
                2020, 1, 1, 11, 00)
            result = plugin._heal_reservation(
                dummy_reservation, list(failed_host.values()))

        pickup_hosts.assert_called_once()
        alloc_destroy.assert_called_once_with('alloc-1')
        self.assertEqual(False, result)

    def test_reallocate_before_start_affinity(self):
        plugin = instance_plugin.VirtualInstancePlugin()
        failed_host = {'id': '1'}
        new_host = {'id': '2'}
        dummy_reservation = {
            'id': 'rsrv-1',
            'resource_type': instances.RESOURCE_TYPE,
            'lease_id': 'lease-1',
            'status': 'pending',
            'vcpus': 2,
            'memory_mb': 1024,
            'disk_gb': 256,
            'aggregate_id': 'agg-1',
            'affinity': True,
            'amount': 3,
            'resource_properties': '',
            'computehost_allocations': [
                {'id': 'alloc-1', 'compute_host_id': failed_host['id'],
                 'reservation_id': 'rsrv-1'},
                {'id': 'alloc-2', 'compute_host_id': failed_host['id'],
                 'reservation_id': 'rsrv-1'},
            ]
        }
        dummy_lease = {
            'name': 'lease-name',
            'start_date': datetime.datetime(2020, 1, 1, 12, 00),
            'end_date': datetime.datetime(2020, 1, 2, 12, 00),
            'trust_id': 'trust-1'
        }
        lease_get = self.patch(db_api, 'lease_get')
        lease_get.return_value = dummy_lease
        pickup_hosts = self.patch(plugin, 'pickup_hosts')
        pickup_hosts.return_value = {'added': [new_host['id']], 'removed': []}
        alloc_update = self.patch(db_api, 'host_allocation_update')

        with mock.patch.object(datetime, 'datetime',
                               mock.Mock(wraps=datetime.datetime)) as patched:
            patched.utcnow.return_value = datetime.datetime(
                2020, 1, 1, 11, 00)
            result = plugin._heal_reservation(
                dummy_reservation, list(failed_host.values()))

        pickup_hosts.assert_called_once()
        update_calls = [mock.call('alloc-1', {'compute_host_id': '2'}),
                        mock.call('alloc-2', {'compute_host_id': '2'})]
        alloc_update.assert_has_calls(update_calls)
        self.assertEqual(True, result)

    def test_reallocate_active_affinity(self):
        plugin = instance_plugin.VirtualInstancePlugin()
        failed_host = {'id': '1',
                       'service_name': 'compute-1',
                       'hypervisor_hostname': 'hypvsr-1'}
        new_host = {'id': '2',
                    'service_name': 'compute-2',
                    'hypervisor_hostname': 'hypvsr-2'}
        dummy_reservation = {
            'id': 'rsrv-1',
            'resource_type': instances.RESOURCE_TYPE,
            'lease_id': 'lease-1',
            'status': 'active',
            'vcpus': 2,
            'memory_mb': 1024,
            'disk_gb': 256,
            'aggregate_id': 'agg-1',
            'affinity': True,
            'amount': 3,
            'resource_properties': '',
            'computehost_allocations': [
                {'id': 'alloc-1', 'compute_host_id': failed_host['id'],
                 'reservation_id': 'rsrv-1'},
                {'id': 'alloc-2', 'compute_host_id': failed_host['id'],
                 'reservation_id': 'rsrv-1'},
            ]
        }
        dummy_lease = {
            'name': 'lease-name',
            'start_date': datetime.datetime(2020, 1, 1, 12, 00),
            'end_date': datetime.datetime(2020, 1, 2, 12, 00),
            'trust_id': 'trust-1'
        }
        reservation_get = self.patch(db_api, 'reservation_get')
        reservation_get.return_value = dummy_reservation
        lease_get = self.patch(db_api, 'lease_get')
        lease_get.return_value = dummy_lease
        host_get = self.patch(db_api, 'host_get')
        host_get.side_effect = [failed_host, new_host]
        fake_pool = mock.MagicMock()
        mock_pool = self.patch(nova, 'ReservationPool')
        mock_pool.return_value = fake_pool
        pickup_hosts = self.patch(plugin, 'pickup_hosts')
        pickup_hosts.return_value = {'added': [new_host['id']], 'removed': []}
        alloc_update = self.patch(db_api, 'host_allocation_update')
        mock_delete_reservation_inventory = self.patch(
            plugin.placement_client, 'delete_reservation_inventory')
        mock_update_reservation_inventory = self.patch(
            plugin.placement_client, 'update_reservation_inventory')

        with mock.patch.object(datetime, 'datetime',
                               mock.Mock(wraps=datetime.datetime)) as patched:
            patched.utcnow.return_value = datetime.datetime(
                2020, 1, 1, 13, 00)
            result = plugin._heal_reservation(
                dummy_reservation, list(failed_host.values()))

        fake_pool.remove_computehost.assert_called_once_with(
            dummy_reservation['aggregate_id'],
            failed_host['service_name'])
        pickup_hosts.assert_called_once()
        update_calls = [mock.call('alloc-1', {'compute_host_id': '2'}),
                        mock.call('alloc-2', {'compute_host_id': '2'})]
        alloc_update.assert_has_calls(update_calls)
        fake_pool.add_computehost.assert_called_once_with(
            dummy_reservation['aggregate_id'],
            new_host['service_name'],
            stay_in=True)
        mock_delete_reservation_inventory.assert_called_once_with(
            'hypvsr-1', 'rsrv-1')
        mock_update_reservation_inventory.assert_called_once_with(
            'hypvsr-2', 'rsrv-1', 2, additional=True)
        self.assertEqual(True, result)

    def test_reallocate_missing_resources_with_affinity(self):
        plugin = instance_plugin.VirtualInstancePlugin()
        failed_host = {'id': '1',
                       'service_name': 'compute-1'}
        dummy_reservation = {
            'id': 'rsrv-1',
            'resource_type': instances.RESOURCE_TYPE,
            'lease_id': 'lease-1',
            'status': 'pending',
            'vcpus': 2,
            'memory_mb': 1024,
            'disk_gb': 256,
            'aggregate_id': 'agg-1',
            'affinity': True,
            'amount': 3,
            'resource_properties': '',
            'computehost_allocations': [
                {'id': 'alloc-1', 'compute_host_id': failed_host['id'],
                 'reservation_id': 'rsrv-1'},
                {'id': 'alloc-2', 'compute_host_id': failed_host['id'],
                 'reservation_id': 'rsrv-1'},
            ]
        }
        dummy_lease = {
            'name': 'lease-name',
            'start_date': datetime.datetime(2020, 1, 1, 12, 00),
            'end_date': datetime.datetime(2020, 1, 2, 12, 00),
            'trust_id': 'trust-1'
        }
        reservation_get = self.patch(db_api, 'reservation_get')
        reservation_get.return_value = dummy_reservation
        lease_get = self.patch(db_api, 'lease_get')
        lease_get.return_value = dummy_lease
        pickup_hosts = self.patch(plugin, 'pickup_hosts')
        pickup_hosts.side_effect = mgr_exceptions.NotEnoughHostsAvailable
        alloc_destroy = self.patch(db_api, 'host_allocation_destroy')

        with mock.patch.object(datetime, 'datetime',
                               mock.Mock(wraps=datetime.datetime)) as patched:
            patched.utcnow.return_value = datetime.datetime(
                2020, 1, 1, 11, 00)
            result = plugin._heal_reservation(
                dummy_reservation, list(failed_host.values()))

        pickup_hosts.assert_called_once()
        destroy_calls = [mock.call('alloc-1'), mock.call('alloc-2')]
        alloc_destroy.assert_has_calls(destroy_calls)
        self.assertEqual(False, result)

    @ddt.data(False, True, None)
    def test_cleanup_resources(self, affinity):
        instance_reservation = {
            'reservation_id': 'reservation-id1',
            'vcpus': 2,
            'memory_mb': 1024,
            'disk_gb': 20,
            'affinity': affinity
        }

        # Set server_group_id according to the affinity value
        server_group_id = 'group-1' if affinity is not None else None
        instance_reservation['server_group_id'] = server_group_id

        mock_nova_client = self.patch(nova, 'NovaClientWrapper')
        mock_nova_client.return_value = mock.MagicMock()
        mock_nova_pool = self.patch(nova, 'ReservationPool')
        mock_nova_pool.return_value = mock.MagicMock()
        plugin = instance_plugin.VirtualInstancePlugin()
        mock_nova = mock.MagicMock()
        type(plugin).nova = mock_nova

        plugin.cleanup_resources(instance_reservation)

        if affinity is not None:
            mock_nova.nova.server_groups.delete.assert_called_once_with(
                'group-1')
        mock_nova.nova.flavors.delete.assert_called_once_with(
            'reservation-id1')



# Copyright (c) 2021 StackHPC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from unittest import mock

import ddt

from blazar import context
from blazar import enforcement
from blazar.enforcement import exceptions
from blazar.enforcement import filters
from blazar import tests

from oslo_config import cfg


def get_fake_host(host_id):
    return {
        'id': host_id,
        'hypervisor_hostname': 'hypvsr1',
        'service_name': 'compute1',
        'vcpus': 4,
        'cpu_info': 'foo',
        'hypervisor_type': 'xen',
        'hypervisor_version': 1,
        'memory_mb': 8192,
        'local_gb': 10,
    }


def get_fake_lease(**kwargs):
    fake_lease = {
        'id': '1',
        'name': 'lease_test',
        'start_date': datetime.datetime(2014, 1, 1, 1, 23),
        'end_date': datetime.datetime(2014, 1, 1, 2, 23),
        'user_id': '111',
        'project_id': '222',
        'trust_id': '35b17138b3644e6aa1318f3099c5be68',
        'reservations': [{'resource_id': '1234',
                          'resource_type': 'virtual:instance'}],
        'events': [],
        'before_end_date': datetime.datetime(2014, 1, 1, 1, 53),
        'action': None,
        'status': None,
        'status_reason': None}

    if kwargs:
        fake_lease.update(kwargs)

    return fake_lease


@ddt.ddt
class MaxLeaseDurationTestCase(tests.TestCase):
    def setUp(self):
        super(MaxLeaseDurationTestCase, self).setUp()

        self.cfg = cfg
        self.region = 'RegionOne'
        filters.all_filters = ['MaxLeaseDurationFilter']

        self.enforcement = enforcement.UsageEnforcement()

        cfg.CONF.set_override(
            'enabled_filters', filters.all_filters, group='enforcement')
        cfg.CONF.set_override('os_region_name', self.region)

        self.enforcement.load_filters()
        cfg.CONF.set_override('max_lease_duration', 3600, group='enforcement')
        self.fake_service_catalog = tests.FakeServiceCatalog([
            dict(
                type='identity', endpoints=[
                    dict(
                        interface='internal', region=self.region,
                        url='https://fakeauth.com')
                ]
            )
        ])

        self.ctx = context.BlazarContext(
            user_id='111', project_id='222',
            service_catalog=self.fake_service_catalog)
        self.set_context(self.ctx)

        self.fake_host_id = '1'
        self.fake_host = {
            'id': self.fake_host_id,
            'hypervisor_hostname': 'hypvsr1',
            'service_name': 'compute1',
            'vcpus': 4,
            'cpu_info': 'foo',
            'hypervisor_type': 'xen',
            'hypervisor_version': 1,
            'memory_mb': 8192,
            'local_gb': 10,
        }

        self.addCleanup(self.cfg.CONF.clear_override, 'enabled_filters',
                        group='enforcement')
        self.addCleanup(self.cfg.CONF.clear_override, 'max_lease_duration',
                        group='enforcement')
        self.addCleanup(self.cfg.CONF.clear_override,
                        'max_lease_duration_exempt_project_ids',
                        group='enforcement')
        self.addCleanup(self.cfg.CONF.clear_override, 'os_region_name')

    def tearDown(self):
        super(MaxLeaseDurationTestCase, self).tearDown()

    def test_check_create_allowed_with_max_lease_duration(self):
        allocation_candidates = {'virtual:instance': [get_fake_host('1')]}
        lease_values = get_fake_lease()
        reservations = list(lease_values['reservations'])

        del lease_values['reservations']
        ctx = context.current()

        self.enforcement.check_create(ctx, lease_values, reservations,
                                      allocation_candidates)

    def test_check_create_denied_beyond_max_lease_duration(self):
        allocation_candidates = {'virtual:instance': [get_fake_host('1')]}
        lease_values = get_fake_lease(
            end_date=datetime.datetime(2014, 1, 1, 2, 24))
        reservations = list(lease_values['reservations'])

        del lease_values['reservations']
        ctx = context.current()

        self.assertRaises(exceptions.MaxLeaseDurationException,
                          self.enforcement.check_create, ctx, lease_values,
                          reservations, allocation_candidates)

    def test_check_update_allowed(self):
        current_allocations = {'virtual:instance': [get_fake_host('1')]}
        lease = get_fake_lease(end_date=datetime.datetime(2014, 1, 1, 2, 22))
        reservations = list(lease['reservations'])

        new_lease_values = get_fake_lease(
            end_date=datetime.datetime(2014, 1, 1, 2, 23))
        new_reservations = list(new_lease_values['reservations'])
        allocation_candidates = {'virtual:instance': [get_fake_host('2')]}

        del new_lease_values['reservations']
        ctx = context.current()

        with mock.patch.object(datetime, 'datetime',
                               mock.Mock(wraps=datetime.datetime)) as patched:
            patched.utcnow.return_value = datetime.datetime(2014, 1, 1, 1, 1)
            self.enforcement.check_update(
                ctx, lease, new_lease_values, current_allocations,
                allocation_candidates, reservations, new_reservations)

    def test_check_update_denied(self):
        current_allocations = {'virtual:instance': [get_fake_host('1')]}
        lease = get_fake_lease(end_date=datetime.datetime(2014, 1, 1, 2, 22))
        reservations = list(lease['reservations'])

        new_lease_values = get_fake_lease(
            end_date=datetime.datetime(2014, 1, 1, 2, 24))
        new_reservations = list(new_lease_values['reservations'])
        allocation_candidates = {'virtual:instance': [get_fake_host('2')]}

        del new_lease_values['reservations']
        ctx = context.current()

        with mock.patch.object(datetime, 'datetime',
                               mock.Mock(wraps=datetime.datetime)) as patched:
            patched.utcnow.return_value = datetime.datetime(2014, 1, 1, 1, 1)
            self.assertRaises(exceptions.MaxLeaseDurationException,
                              self.enforcement.check_update, ctx, lease,
                              new_lease_values, current_allocations,
                              allocation_candidates, reservations,
                              new_reservations)

    def test_check_update_active_lease_allowed(self):
        current_allocations = {'virtual:instance': [get_fake_host('1')]}
        lease = get_fake_lease(end_date=datetime.datetime(2014, 1, 1, 1, 53))
        reservations = list(lease['reservations'])

        new_lease_values = get_fake_lease()
        new_reservations = list(new_lease_values['reservations'])
        allocation_candidates = {'virtual:instance': [get_fake_host('2')]}

        del new_lease_values['reservations']
        ctx = context.current()

        with mock.patch.object(datetime, 'datetime',
                               mock.Mock(wraps=datetime.datetime)) as patched:
            patched.utcnow.return_value = datetime.datetime(2014, 1, 1, 1, 50)
            self.enforcement.check_update(
                ctx, lease, new_lease_values, current_allocations,
                allocation_candidates, reservations, new_reservations)

    def test_check_create_exempt(self):
        cfg.CONF.set_override('max_lease_duration_exempt_project_ids', ['222'],
                              group='enforcement')
        allocation_candidates = {'virtual:instance': [get_fake_host('1')]}
        lease_values = get_fake_lease(
            end_date=datetime.datetime(2014, 1, 1, 2, 24))
        reservations = list(lease_values['reservations'])

        del lease_values['reservations']
        ctx = context.current()

        self.enforcement.check_create(ctx, lease_values, reservations,
                                      allocation_candidates)

    def test_check_update_exempt(self):
        cfg.CONF.set_override('max_lease_duration_exempt_project_ids', ['222'],
                              group='enforcement')
        current_allocations = {'virtual:instance': [get_fake_host('1')]}
        lease = get_fake_lease(end_date=datetime.datetime(2014, 1, 1, 2, 22))
        reservations = list(lease['reservations'])

        new_lease_values = get_fake_lease(
            end_date=datetime.datetime(2014, 1, 1, 2, 24))
        new_reservations = list(new_lease_values['reservations'])
        allocation_candidates = {'virtual:instance': [get_fake_host('2')]}

        del new_lease_values['reservations']
        ctx = context.current()

        with mock.patch.object(datetime, 'datetime',
                               mock.Mock(wraps=datetime.datetime)) as patched:
            patched.utcnow.return_value = datetime.datetime(2014, 1, 1, 1, 1)
            self.enforcement.check_update(
                ctx, lease, new_lease_values, current_allocations,
                allocation_candidates, reservations, new_reservations)

    def test_on_end(self):
        allocations = {'virtual:instance': [get_fake_host('1')]}
        lease = get_fake_lease()
        ctx = context.current()

        self.enforcement.on_end(ctx, lease, allocations)



# Copyright (c) 2013 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ddt
import flask
from oslo_utils import uuidutils
from testtools import matchers

from oslo_middleware import request_id as id

from blazar.api import context as api_context
from blazar.api.v1 import api_version_request
from blazar.api.v1.oshosts import service as service_api
from blazar.api.v1.oshosts import v1_0 as hosts_api_v1_0
from blazar.api.v1 import request_id
from blazar.api.v1 import request_log
from blazar import context
from blazar import tests


def make_app():
    """App builder (wsgi).

    Entry point for Blazar REST API server.
    """
    app = flask.Flask('blazar.api')

    app.register_blueprint(hosts_api_v1_0.rest, url_prefix='/v1')
    app.wsgi_app = request_id.BlazarReqIdMiddleware(app.wsgi_app)
    app.wsgi_app = request_log.RequestLog(app.wsgi_app)

    return app


def fake_computehost(**kw):
    return {
        'id': kw.get('id', '1'),
        'hypervisor_hostname': kw.get('hypervisor_hostname', 'host01'),
        'hypervisor_type': kw.get('hypervisor_type', 'QEMU'),
        'vcpus': kw.get('vcpus', 1),
        'hypervisor_version': kw.get('hypervisor_version', 1000000),
        'trust_id': kw.get('trust_id',
                           '35b17138-b364-4e6a-a131-8f3099c5be68'),
        'memory_mb': kw.get('memory_mb', 8192),
        'local_gb': kw.get('local_gb', 50),
        'cpu_info': kw.get('cpu_info',
                           "{\"vendor\": \"Intel\", \"model\": \"qemu32\", "
                           "\"arch\": \"x86_64\", \"features\": [],"
                           " \"topology\": {\"cores\": 1}}",
                           ),
        'extra_capas': kw.get('extra_capas',
                              {'vgpus': 2, 'fruits': 'bananas'})
    }


def fake_computehost_request_body(include=None, **kw):
    computehost_body = fake_computehost(**kw)
    computehost_body['name'] = kw.get('name',
                                      computehost_body['hypervisor_hostname'])
    default_include = set(['name', 'extra_capas'])
    include = include or set()
    include |= default_include
    return dict((key, computehost_body[key])
                for key in computehost_body if key in include)


@ddt.ddt
class OsHostAPITestCase(tests.TestCase):

    def setUp(self):
        super(OsHostAPITestCase, self).setUp()
        self.app = make_app()
        self.headers = {'Accept': 'application/json',
                        'OpenStack-API-Version': 'reservation 1.0'}
        self.host_id = str('1')
        self.mock_ctx = self.patch(api_context, 'ctx_from_headers')
        self.mock_ctx.return_value = context.BlazarContext(
            user_id='fake', project_id='fake', roles=['member'])
        self.get_computehosts = self.patch(service_api.API,
                                           'get_computehosts')
        self.create_computehost = self.patch(service_api.API,
                                             'create_computehost')
        self.get_computehost = self.patch(service_api.API, 'get_computehost')
        self.update_computehost = self.patch(service_api.API,
                                             'update_computehost')
        self.delete_computehost = self.patch(service_api.API,
                                             'delete_computehost')
        self.list_allocations = self.patch(service_api.API,
                                           'list_allocations')
        self.get_allocations = self.patch(service_api.API, 'get_allocations')
        self.list_resource_properties = self.patch(service_api.API,
                                                   'list_resource_properties')
        self.update_resource_property = self.patch(service_api.API,
                                                   'update_resource_property')

    def _assert_response(self, actual_resp, expected_status_code,
                         expected_resp_body, key='host',
                         expected_api_version='reservation 1.0'):
        res_id = actual_resp.headers.get(id.HTTP_RESP_HEADER_REQUEST_ID)
        api_version = actual_resp.headers.get(
            api_version_request.API_VERSION_REQUEST_HEADER)
        self.assertIn(id.HTTP_RESP_HEADER_REQUEST_ID,
                      actual_resp.headers)
        self.assertIn(api_version_request.API_VERSION_REQUEST_HEADER,
                      actual_resp.headers)
        self.assertIn(api_version_request.VARY_HEADER, actual_resp.headers)
        self.assertThat(res_id, matchers.StartsWith('req-'))
        self.assertEqual(expected_status_code, actual_resp.status_code)
        self.assertEqual(expected_resp_body, actual_resp.get_json()[key])
        self.assertEqual(expected_api_version, api_version)
        self.assertEqual('OpenStack-API-Version', actual_resp.headers.get(
            api_version_request.VARY_HEADER))

    def test_list(self):
        with self.app.test_client() as c:
            self.get_computehosts.return_value = []
            res = c.get('/v1', headers=self.headers)
            self._assert_response(res, 200, [], key='hosts')

    def test_list_with_non_acceptable_version(self):
        headers = {'Accept': 'application/json',
                   'OpenStack-API-Version': 'reservation 1.2'}
        with self.app.test_client() as c:
            res = c.get('/v1', headers=headers)
            self.assertEqual(406, res.status_code)

    def test_create(self):
        with self.app.test_client() as c:
            self.create_computehost.return_value = fake_computehost(
                id=self.host_id)
            res = c.post('/v1', json=fake_computehost_request_body(
                id=self.host_id), headers=self.headers)
            self._assert_response(res, 201, fake_computehost(
                id=self.host_id))

    def test_create_with_bad_api_version(self):
        headers = {'Accept': 'application/json',
                   'OpenStack-API-Version': 'reservation 1.a'}
        with self.app.test_client() as c:
            res = c.post('/v1', json=fake_computehost_request_body(
                id=self.host_id), headers=headers)
            self.assertEqual(400, res.status_code)

    def test_get(self):
        with self.app.test_client() as c:
            self.get_computehost.return_value = fake_computehost(
                id=self.host_id)
            res = c.get('/v1/{0}'.format(self.host_id), headers=self.headers)
            self._assert_response(res, 200, fake_computehost(id=self.host_id))

    def test_get_with_latest_api_version(self):
        headers = {'Accept': 'application/json',
                   'OpenStack-API-Version': 'reservation latest'}
        with self.app.test_client() as c:
            self.get_computehost.return_value = fake_computehost(
                id=self.host_id)
            res = c.get('/v1/{0}'.format(self.host_id), headers=headers)
            self._assert_response(res, 200, fake_computehost(id=self.host_id),
                                  expected_api_version='reservation 1.0')

    def test_update(self):
        headers = {'Accept': 'application/json'}
        with self.app.test_client() as c:
            self.fake_computehost = fake_computehost(id=self.host_id,
                                                     name='updated')
            self.fake_computehost_body = fake_computehost_request_body(
                id=self.host_id,
                name='updated'
            )
            self.update_computehost.return_value = self.fake_computehost

            res = c.put('/v1/{0}'.format(self.host_id),
                        json=self.fake_computehost_body, headers=headers)
            self._assert_response(res, 200, self.fake_computehost, 'host')

    def test_update_with_no_service_type_in_header(self):
        headers = {'Accept': 'application/json',
                   'OpenStack-API-Version': '1.0'}
        with self.app.test_client() as c:
            self.fake_computehost = fake_computehost(id=self.host_id,
                                                     name='updated')
            self.fake_computehost_body = fake_computehost_request_body(
                id=self.host_id,
                name='updated'
            )
            self.update_computehost.return_value = self.fake_computehost

            res = c.put('/v1/{0}'.format(self.host_id),
                        json=self.fake_computehost_body, headers=headers)
            self._assert_response(res, 200, self.fake_computehost, 'host')

    def test_delete(self):
        with self.app.test_client() as c:
            self.get_computehosts.return_value = fake_computehost(
                id=self.host_id)

            res = c.delete('/v1/{0}'.format(self.host_id),
                           headers=self.headers)
            res_id = res.headers.get(id.HTTP_RESP_HEADER_REQUEST_ID)
            self.assertEqual(204, res.status_code)
            self.assertIn(id.HTTP_RESP_HEADER_REQUEST_ID, res.headers)
            self.assertThat(res_id, matchers.StartsWith('req-'))

    def test_allocation_list(self):
        with self.app.test_client() as c:
            self.list_allocations.return_value = []
            res = c.get('/v1/allocations', headers=self.headers)
            self._assert_response(res, 200, [], key='allocations')

    def test_allocation_get(self):
        with self.app.test_client() as c:
            self.get_allocations.return_value = {}
            res = c.get('/v1/{0}/allocation'.format(self.host_id),
                        headers=self.headers)
            self._assert_response(res, 200, {}, key='allocation')

    @ddt.data({'lease_id': str(uuidutils.generate_uuid()),
               'reservation_id': str(uuidutils.generate_uuid())})
    def test_allocation_list_with_query_params(self, query_params):
        with self.app.test_client() as c:
            res = c.get('/v1/allocations?{0}'.format(query_params),
                        headers=self.headers)
            self._assert_response(res, 200, {}, key='allocations')

    @ddt.data({'lease_id': str(uuidutils.generate_uuid()),
               'reservation_id': str(uuidutils.generate_uuid())})
    def test_allocation_get_with_query_params(self, query_params):
        with self.app.test_client() as c:
            res = c.get('/v1/{0}/allocation?{1}'.format(
                self.host_id, query_params), headers=self.headers)
            self._assert_response(res, 200, {}, key='allocation')

    def test_resource_properties_list(self):
        with self.app.test_client() as c:
            self.list_resource_properties.return_value = []
            res = c.get('/v1/properties', headers=self.headers)
            self._assert_response(res, 200, [], key='resource_properties')

    def test_resource_property_update(self):
        resource_property = 'fake_property'
        resource_property_body = {'private': True}

        with self.app.test_client() as c:

            res = c.patch('/v1/properties/{0}'.format(resource_property),
                          json=resource_property_body,
                          headers=self.headers)
            self._assert_response(res, 200, {}, 'resource_property')



# -*- encoding: utf-8 -*-
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests for API /leases/ methods
"""
from oslo_utils import uuidutils

from blazar.tests import api
from blazar.utils import trusts


def fake_lease(**kw):
    return {
        'id': kw.get('id', '2bb8720a-0873-4d97-babf-0d906851a1eb'),
        'name': kw.get('name', 'lease_test'),
        'start_date': kw.get('start_date', '2014-01-01 01:23'),
        'end_date': kw.get('end_date', '2014-02-01 13:37'),
        'trust_id': kw.get('trust_id',
                           '35b17138b3644e6aa1318f3099c5be68'),
        'user_id': kw.get('user_id', api.FAKE_USER),
        'project_id': kw.get('project_id', api.FAKE_PROJECT),
        'reservations': kw.get('reservations', [
            {
                'resource_id': '1234',
                'resource_type': 'virtual:instance'
            }
        ]),
        'events': kw.get('events', []),
        'status': kw.get('status', 'ACTIVE'),
    }


def fake_lease_request_body(exclude=[], **kw):
    exclude.append('id')
    exclude.append('trust_id')
    exclude.append('user_id')
    exclude.append('project_id')
    exclude.append('status')
    lease_body = fake_lease(**kw)
    return dict((key, lease_body[key])
                for key in lease_body if key not in exclude)


def fake_trust(id=fake_lease()['trust_id']):
    return type('Trust', (), {
        'id': id,
    })


class TestIncorrectLeaseFromRPC(api.APITest):

    def setUp(self):
        super(TestIncorrectLeaseFromRPC, self).setUp()

        self.path = '/leases'
        self.patch(
            self.rpcapi, 'list_leases').return_value = [fake_lease(id=1)]

    def test_bad_list(self):
        expected = {
            'error_code': 400,
            'error_message': "Invalid input for field/attribute id. "
                             "Value: '1'. Value should be UUID format",
            'error_name': 400
        }
        response = self.get_json(self.path, expect_errors=True)
        self.assertEqual(400, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected, response.json)


class TestListLeases(api.APITest):

    def setUp(self):
        super(TestListLeases, self).setUp()

        self.fake_lease = fake_lease()
        self.path = '/leases'
        self.patch(self.rpcapi, 'list_leases').return_value = []

    def test_empty(self):
        response = self.get_json(self.path)
        self.assertEqual([], response)

    def test_one(self):
        self.patch(self.rpcapi, 'list_leases').return_value = [self.fake_lease]
        response = self.get_json(self.path)
        self.assertEqual([self.fake_lease], response)

    def test_multiple(self):
        id1 = str(uuidutils.generate_uuid())
        id2 = str(uuidutils.generate_uuid())
        self.patch(
            self.rpcapi, 'list_leases').return_value = [
                fake_lease(id=id1),
                fake_lease(id=id2)
            ]
        response = self.get_json(self.path)
        self.assertEqual([fake_lease(id=id1), fake_lease(id=id2)], response)

    def test_rpc_exception_list(self):
        def fake_list_leases(*args, **kwargs):
            raise Exception("Nah...")
        expected = {
            'error_code': 500,
            'error_message': "Nah...",
            'error_name': 500
        }
        self.patch(
            self.rpcapi, 'list_leases').side_effect = fake_list_leases
        response = self.get_json(self.path, expect_errors=True)
        self.assertEqual(500, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected, response.json)


class TestShowLease(api.APITest):

    def setUp(self):
        super(TestShowLease, self).setUp()

        self.id1 = str(uuidutils.generate_uuid())
        self.fake_lease = fake_lease(id=self.id1)
        self.path = '/leases/{0}'.format(self.id1)
        self.patch(self.rpcapi, 'get_lease').return_value = self.fake_lease

    def test_one(self):
        response = self.get_json(self.path)
        self.assertEqual(self.fake_lease, response)

    def test_empty(self):
        expected = {
            'error_code': 404,
            'error_message': "not found",
            'error_name': 404
        }
        self.patch(self.rpcapi, 'get_lease').return_value = None
        response = self.get_json(self.path, expect_errors=True)
        self.assertEqual(404, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected["error_name"], response.json["error_name"])
        self.assertIn(expected["error_message"],
                      response.json["error_message"])
        self.assertEqual(expected["error_code"], response.json["error_code"])

    def test_rpc_exception_get(self):
        def fake_get_lease(*args, **kwargs):
            raise Exception("Nah...")
        expected = {
            'error_code': 500,
            'error_message': "Nah...",
            'error_name': 500
        }
        self.patch(
            self.rpcapi, 'get_lease').side_effect = fake_get_lease
        response = self.get_json(self.path, expect_errors=True)
        self.assertEqual(500, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected, response.json)


class TestCreateLease(api.APITest):

    def setUp(self):
        super(TestCreateLease, self).setUp()

        self.id1 = str(uuidutils.generate_uuid())
        self.fake_lease = fake_lease(id=self.id1)
        self.fake_lease_body = fake_lease_request_body(id=self.id1)
        self.path = '/leases'
        self.patch(self.rpcapi, 'create_lease').return_value = self.fake_lease

        self.trusts = trusts
        self.patch(self.trusts, 'create_trust').return_value = fake_trust()

    def test_create_one(self):
        response = self.post_json(self.path, self.fake_lease_body)
        self.assertEqual(201, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(self.fake_lease, response.json)

    def test_create_wrong_attr(self):
        expected = {
            "error_name": 400,
            "error_message": "Invalid input for field/attribute name.",
            "error_code": 400
        }

        response = self.post_json(self.path, fake_lease_request_body(name=1),
                                  expect_errors=True)
        self.assertEqual(400, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected["error_name"], response.json["error_name"])
        self.assertIn(expected["error_message"],
                      response.json["error_message"])
        self.assertEqual(expected["error_code"], response.json["error_code"])

    def test_create_with_empty_body(self):
        expected = {
            "error_name": 500,
            "error_message": "'NoneType' object has no attribute 'as_dict'",
            "error_code": 500
        }

        response = self.post_json(self.path, None, expect_errors=True)
        self.assertEqual(500, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected, response.json)

    def test_empty_response(self):
        expected = {
            'error_code': 500,
            'error_message': "Lease can't be created",
            'error_name': 500
        }
        self.patch(self.rpcapi, 'create_lease').return_value = None
        response = self.post_json(self.path, self.fake_lease_body,
                                  expect_errors=True)
        self.assertEqual(500, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected, response.json)

    def test_rpc_exception_create(self):
        def fake_create_lease(*args, **kwargs):
            raise Exception("Nah...")
        expected = {
            'error_code': 500,
            'error_message': "Nah...",
            'error_name': 500
        }
        self.patch(
            self.rpcapi, 'create_lease').side_effect = fake_create_lease
        response = self.post_json(self.path, self.fake_lease_body,
                                  expect_errors=True)
        self.assertEqual(500, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected, response.json)


class TestUpdateLease(api.APITest):

    def setUp(self):
        super(TestUpdateLease, self).setUp()

        self.id1 = str(uuidutils.generate_uuid())
        self.fake_lease = fake_lease(id=self.id1, name='updated')
        self.fake_lease_body = fake_lease_request_body(
            exclude=['reservations', 'events'],
            id=self.id1,
            name='updated'
        )
        self.path = '/leases/{0}'.format(self.id1)
        self.patch(self.rpcapi, 'update_lease').return_value = self.fake_lease

    def test_update_one(self):
        response = self.put_json(self.path, self.fake_lease_body)
        self.assertEqual(200, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(self.fake_lease, response.json)

    def test_update_one_with_extra_attrs(self):
        expected = {
            "error_name": 500,
            "error_message": "Only name changing, dates and before "
                             "end notifications may be proceeded.",
            "error_code": 500
        }

        response = self.put_json(self.path, fake_lease_request_body(name='a'),
                                 expect_errors=True)
        self.assertEqual(500, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected, response.json)

    def test_update_with_empty_body(self):
        expected = {
            "error_name": 500,
            "error_message": "'NoneType' object has no attribute 'as_dict'",
            "error_code": 500
        }

        response = self.put_json(self.path, None, expect_errors=True)
        self.assertEqual(500, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected, response.json)

    def test_empty_response(self):
        expected = {
            'error_code': 404,
            'error_message': "not found",
            'error_name': 404
        }
        self.patch(self.rpcapi, 'update_lease').return_value = None
        response = self.put_json(self.path, self.fake_lease_body,
                                 expect_errors=True)
        self.assertEqual(404, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected["error_name"], response.json["error_name"])
        self.assertIn(expected["error_message"],
                      response.json["error_message"])
        self.assertEqual(expected["error_code"], response.json["error_code"])

    def test_rpc_exception_update(self):
        def fake_update_lease(*args, **kwargs):
            raise Exception("Nah...")
        expected = {
            'error_code': 500,
            'error_message': "Nah...",
            'error_name': 500
        }
        self.patch(
            self.rpcapi, 'update_lease').side_effect = fake_update_lease
        response = self.put_json(self.path, self.fake_lease_body,
                                 expect_errors=True)
        self.assertEqual(500, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected, response.json)


class TestDeleteLease(api.APITest):

    def setUp(self):
        super(TestDeleteLease, self).setUp()

        self.id1 = str(uuidutils.generate_uuid())
        self.path = '/leases/{0}'.format(self.id1)
        self.patch(self.rpcapi, 'delete_lease')

    def test_delete_one(self):
        response = self.delete(self.path)
        self.assertEqual(204, response.status_int)
        self.assertIsNone(response.content_type)
        self.assertEqual(b'', response.body)

    def test_delete_not_existing_lease(self):
        def fake_delete_lease(*args, **kwargs):
            raise TypeError("Nah...")
        expected = {
            'error_code': 404,
            'error_message': "not found",
            'error_name': 404
        }
        self.patch(
            self.rpcapi, 'delete_lease').side_effect = fake_delete_lease
        response = self.delete(self.path, expect_errors=True)
        self.assertEqual(404, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected["error_name"], response.json["error_name"])
        self.assertIn(expected["error_message"],
                      response.json["error_message"])
        self.assertEqual(expected["error_code"], response.json["error_code"])

    def test_rpc_exception_delete(self):
        def fake_delete_lease(*args, **kwargs):
            raise Exception("Nah...")
        expected = {
            'error_code': 500,
            'error_message': "Nah...",
            'error_name': 500
        }
        self.patch(
            self.rpcapi, 'delete_lease').side_effect = fake_delete_lease
        response = self.delete(self.path, expect_errors=True)
        self.assertEqual(500, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected, response.json)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines interface for DB access.

Functions in this module are imported into the blazar.db namespace. Call these
functions from blazar.db namespace, not the blazar.db.api namespace.

All functions in this module return objects that implement a dictionary-like
interface.

**Related Flags**

:db_backend:  string to lookup in the list of LazyPluggable backends.
              `sqlalchemy` is the only supported backend right now.

:sql_connection:  string specifying the sqlalchemy connection to use, like:
                  `sqlite:///var/lib/blazar/blazar.sqlite`.

"""

from oslo_config import cfg
from oslo_db import api as db_api
from oslo_db import options as db_options
from oslo_log import log as logging


_BACKEND_MAPPING = {
    'sqlalchemy': 'blazar.db.sqlalchemy.api',
}

db_options.set_defaults(cfg.CONF)
IMPL = db_api.DBAPI(cfg.CONF.database.backend,
                    backend_mapping=_BACKEND_MAPPING)
LOG = logging.getLogger(__name__)

# The maximum value a signed INT type may have
DB_MAX_INT = 0x7FFFFFFF


def get_instance():
    """Return a DB API instance."""
    return IMPL


def setup_db():
    """Set up database, create tables, etc.

    Return True on success, False otherwise
    """
    return IMPL.setup_db()


def drop_db():
    """Drop database.

    Return True on success, False otherwise
    """
    return IMPL.drop_db()


# Helpers for building constraints / equality checks


def constraint(**conditions):
    """Return a constraint object suitable for use with some updates."""
    return IMPL.constraint(**conditions)


def equal_any(*values):
    """Return an equality condition object suitable for use in a constraint.

    Equal_any conditions require that a model object's attribute equal any
    one of the given values.
    """
    return IMPL.equal_any(*values)


def not_equal(*values):
    """Return an inequality condition object suitable for use in a constraint.

    Not_equal conditions require that a model object's attribute differs from
    all of the given values.
    """
    return IMPL.not_equal(*values)


def to_dict(func):
    def decorator(*args, **kwargs):
        res = func(*args, **kwargs)

        if isinstance(res, list):
            return [item.to_dict() for item in res]

        if res:
            return res.to_dict()
        else:
            return None

    return decorator


# Reservation

def reservation_create(reservation_values):
    """Create a reservation from the values."""
    return IMPL.reservation_create(reservation_values)


@to_dict
def reservation_get_all_by_lease_id(lease_id):
    """Return all reservations belongs to specific lease."""
    return IMPL.reservation_get_all_by_lease_id(lease_id)


@to_dict
def reservation_get_all_by_values(**kwargs):
    """Returns all entries filtered by col=value."""
    return IMPL.reservation_get_all_by_values(**kwargs)


@to_dict
def reservation_get(reservation_id):
    """Return specific reservation."""
    return IMPL.reservation_get(reservation_id)


def reservation_destroy(reservation_id):
    """Delete specific reservation."""
    IMPL.reservation_destroy(reservation_id)


def reservation_update(reservation_id, reservation_values):
    """Update reservation."""
    IMPL.reservation_update(reservation_id, reservation_values)


# Lease

def lease_create(lease_values):
    """Create a lease from values."""
    return IMPL.lease_create(lease_values)


@to_dict
def lease_get_all():
    """Return all leases."""
    return IMPL.lease_get_all()


@to_dict
def lease_get_all_by_project(project_id):
    """Return all leases in specific project."""
    return IMPL.lease_get_all_by_project(project_id)


@to_dict
def lease_get_all_by_user(user_id):
    """Return all leases belongs to specific user."""
    return IMPL.lease_get_all_by_user(user_id)


@to_dict
def lease_get(lease_id):
    """Return lease."""
    return IMPL.lease_get(lease_id)


@to_dict
def lease_list(project_id=None):
    """Return a list of all existing leases."""
    return IMPL.lease_list(project_id)


def lease_destroy(lease_id):
    """Delete lease or raise if not exists."""
    IMPL.lease_destroy(lease_id)


def lease_update(lease_id, lease_values):
    """Update lease or raise if not exists."""
    IMPL.lease_update(lease_id, lease_values)


# Events

@to_dict
def event_create(event_values):
    """Create an event from values."""
    return IMPL.event_create(event_values)


@to_dict
def event_get_all():
    """Return all events."""
    return IMPL.event_get_all()


@to_dict
def event_get(event_id):
    """Return a specific event."""
    return IMPL.event_get(event_id)


@to_dict
def event_get_first_sorted_by_filters(sort_key, sort_dir, filters):
    """Return instances sorted by param."""
    return IMPL.event_get_first_sorted_by_filters(sort_key, sort_dir,
                                                  filters)


@to_dict
def event_get_all_sorted_by_filters(sort_key, sort_dir, filters):
    """Return instances sorted by param."""
    return IMPL.event_get_all_sorted_by_filters(sort_key, sort_dir,
                                                filters)


def event_destroy(event_id):
    """Delete event or raise if not exists."""
    IMPL.event_destroy(event_id)


def event_update(event_id, event_values):
    """Update event or raise if not exists."""
    IMPL.event_update(event_id, event_values)


# Host reservations

def host_reservation_create(host_reservation_values):
    """Create a host reservation from the values."""
    return IMPL.host_reservation_create(host_reservation_values)


@to_dict
def host_reservation_get_by_reservation_id(reservation_id):
    """Return host reservation belonging to specific reservation."""
    return IMPL.host_reservation_get_by_reservation_id(reservation_id)


@to_dict
def host_reservation_get(host_reservation_id):
    """Return specific host reservation."""
    return IMPL.host_reservation_get(host_reservation_id)


@to_dict
def host_reservation_get_all():
    """Return all hosts reservations."""
    return IMPL.host_reservation_get_all()


def host_reservation_destroy(host_reservation_id):
    """Delete specific host reservation."""
    IMPL.host_reservation_destroy(host_reservation_id)


def host_reservation_update(host_reservation_id,
                            host_reservation_values):
    """Update host reservation."""
    IMPL.host_reservation_update(host_reservation_id,
                                 host_reservation_values)


# Instance reservation

def instance_reservation_create(instance_reservation_values):
    """Create a instance reservation from the values."""
    return IMPL.instance_reservation_create(instance_reservation_values)


def instance_reservation_get(instance_reservation_id):
    """Return specific instance reservation."""
    return IMPL.instance_reservation_get(instance_reservation_id)


def instance_reservation_update(instance_reservation_id,
                                instance_reservation_values):
    """Update instance reservation."""
    return IMPL.instance_reservation_update(instance_reservation_id,
                                            instance_reservation_values)


def instance_reservation_destroy(instance_reservation_id):
    """Delete specific instance reservation."""
    return IMPL.instance_reservation_destroy(instance_reservation_id)


# Allocation

def host_allocation_create(allocation_values):
    """Create an allocation from the values."""
    return IMPL.host_allocation_create(allocation_values)


@to_dict
def host_allocation_get_all_by_values(**kwargs):
    """Returns all entries filtered by col=value."""
    return IMPL.host_allocation_get_all_by_values(**kwargs)


# TODO(frossigneux) get methods


def host_allocation_destroy(allocation_id):
    """Delete specific allocation."""
    IMPL.host_allocation_destroy(allocation_id)


def host_allocation_update(allocation_id, allocation_values):
    """Update allocation."""
    IMPL.host_allocation_update(allocation_id, allocation_values)


# Compute Hosts

def host_create(values):
    """Create a Compute host from the values."""
    return IMPL.host_create(values)


@to_dict
def host_get(host_id):
    """Return a specific Compute host."""
    return IMPL.host_get(host_id)


@to_dict
def host_list():
    """Return a list of events."""
    return IMPL.host_list()


@to_dict
def host_get_all_by_filters(filters):
    """Returns Compute hosts filtered by name of the field."""
    return IMPL.host_get_all_by_filters(filters)


@to_dict
def host_get_all_by_queries(queries):
    """Returns hosts filtered by an array of queries."""
    return IMPL.host_get_all_by_queries(queries)


@to_dict
def reservable_host_get_all_by_queries(queries):
    """Returns reservable hosts filtered by an array of queries."""
    return IMPL.reservable_host_get_all_by_queries(queries)


@to_dict
def unreservable_host_get_all_by_queries(queries):
    """Returns unreservable hosts filtered by an array of queries."""
    return IMPL.unreservable_host_get_all_by_queries(queries)


def host_destroy(host_id):
    """Delete specific Compute host."""
    IMPL.host_destroy(host_id)


def host_update(host_id, values):
    """Update Compute host."""
    IMPL.host_update(host_id, values)


# ComputeHostExtraCapabilities

def host_extra_capability_create(values):
    """Create a Host ExtraCapability from the values."""
    return IMPL.host_extra_capability_create(values)


def host_extra_capability_get(host_extra_capability_id):
    """Return a specific Host Extracapability."""
    return IMPL.host_extra_capability_get(host_extra_capability_id)


def host_extra_capability_get_all_per_host(host_id):
    """Return all extra_capabilities belonging to a specific Compute host."""
    return IMPL.host_extra_capability_get_all_per_host(host_id)


def host_extra_capability_destroy(host_extra_capability_id):
    """Delete specific host ExtraCapability."""
    IMPL.host_extra_capability_destroy(host_extra_capability_id)


def host_extra_capability_update(host_extra_capability_id, values):
    """Update specific host ExtraCapability."""
    IMPL.host_extra_capability_update(host_extra_capability_id, values)


def host_extra_capability_get_all_per_name(host_id,
                                           extra_capability_name):
    return IMPL.host_extra_capability_get_all_per_name(host_id,
                                                       extra_capability_name)


# Host matching

def host_get_all_by_queries_including_extracapabilities(queries):
    """Returns hosts filtered by an array of queries."""
    return IMPL.host_get_all_by_queries_including_extracapabilities(queries)


# ComputeHostResourceInventory

def host_resource_inventory_create(values):
    """Create a host resource inventory from the values."""
    return IMPL.host_resource_inventory_create(values)


def host_resource_inventory_get_all_per_host(host_id):
    """Return all resource inventories belonging to a specific Compute host."""
    return IMPL.host_resource_inventory_get_all_per_host(host_id)


# ComputeHostTrait

def host_trait_create(values):
    """Create a host trait from the values."""
    return IMPL.host_trait_create(values)


# FloatingIP reservation

def fip_reservation_create(fip_reservation_values):
    """Create a floating IP reservation from the values."""
    return IMPL.fip_reservation_create(fip_reservation_values)


@to_dict
def fip_reservation_get(fip_reservation_id):
    """Return specific floating IP reservation."""
    return IMPL.fip_reservation_get(fip_reservation_id)


def fip_reservation_update(fip_reservation_id, fip_reservation_values):
    """Update floating IP reservation."""
    return IMPL.fip_reservation_update(fip_reservation_id,
                                       fip_reservation_values)


def fip_reservation_destroy(fip_reservation_id):
    """Delete specific floating ip reservation."""
    return IMPL.fip_reservation_destroy(fip_reservation_id)


# Required FloatingIP

def required_fip_create(required_fip_values):
    """Create a required FIP address from the values."""
    return IMPL.required_fip_create(required_fip_values)


@to_dict
def required_fip_get(required_fip_id):
    """Return specific required FIP."""
    return IMPL.required_fip_get(required_fip_id)


def required_fip_update(required_fip_id, required_fip_values):
    """Update required FIP."""
    return IMPL.required_fip_update(required_fip_id,
                                    required_fip_values)


def required_fip_destroy(required_fip_id):
    """Delete specific required FIP."""
    return IMPL.required_fip_destroy(required_fip_id)


def required_fip_destroy_by_fip_reservation_id(fip_reservation_id):
    """Delete all required FIPs for a floating IP reservation."""
    return IMPL.required_fip_destroy_by_fip_reservation_id(fip_reservation_id)


# FloatingIP Allocation

def fip_allocation_create(allocation_values):
    """Create a floating ip allocation from the values."""
    return IMPL.fip_allocation_create(allocation_values)


@to_dict
def fip_allocation_get_all_by_values(**kwargs):
    """Returns all entries filtered by col=value."""
    return IMPL.fip_allocation_get_all_by_values(**kwargs)


def fip_allocation_destroy(allocation_id):
    """Delete specific floating ip allocation."""
    IMPL.fip_allocation_destroy(allocation_id)


def fip_allocation_update(allocation_id, allocation_values):
    """Update floating ip allocation."""
    IMPL.fip_allocation_update(allocation_id, allocation_values)


# Floating ip

def floatingip_create(values):
    """Create a floating ip from the values."""
    return IMPL.floatingip_create(values)


@to_dict
def floatingip_get(floatingip_id):
    """Return a specific floating ip."""
    return IMPL.floatingip_get(floatingip_id)


@to_dict
def floatingip_list():
    """Return a list of floating ip."""
    return IMPL.floatingip_list()


@to_dict
def reservable_fip_get_all_by_queries(queries):
    """Returns reservable fips filtered by an array of queries."""
    return IMPL.reservable_fip_get_all_by_queries(queries)


def floatingip_destroy(floatingip_id):
    """Delete specific floating ip."""
    IMPL.floatingip_destroy(floatingip_id)


# Resource Properties

def resource_properties_list(resource_type):
    return IMPL.resource_properties_list(resource_type)


def resource_property_update(resource_type, property_name, values):
    return IMPL.resource_property_update(resource_type, property_name, values)


def resource_property_create(values):
    return IMPL.resource_property_create(values)



# Copyright 2017 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Delete unused status related fields

Revision ID: e66f199a5414
Revises: 97d2cad1504e
Create Date: 2017-12-18 15:33:09.217659

"""

# revision identifiers, used by Alembic.
revision = 'e66f199a5414'
down_revision = '97d2cad1504e'

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import mysql


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('computehost_reservations', 'status')
    op.drop_column('leases', 'action')
    op.drop_column('leases', 'status_reason')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('leases', sa.Column('status_reason',
                                      mysql.VARCHAR(length=255),
                                      nullable=True))
    op.add_column('leases', sa.Column('action',
                                      mysql.VARCHAR(length=255),
                                      nullable=True))
    op.add_column('computehost_reservations',
                  sa.Column('status',
                            mysql.VARCHAR(length=13),
                            nullable=True))
    # ### end Alembic commands ###



# Copyright (c) 2014 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

from oslo_config import cfg
from oslo_log import log as logging
from pecan import rest

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class BaseController(rest.RestController, metaclass=abc.ABCMeta):

    """Mandatory API method name."""
    name = None

    """Optional extra routes to add.
    Dict of key/value pairs, where :
        key : API method name redirect (public URL)
        value : redirect target (can be None for routing to HTTP 404)
    """
    extra_routes = {}

    @abc.abstractmethod
    def get_one(self, resource_id):
        """Get a single resource."""
        pass

    @abc.abstractmethod
    def get_all(self):
        """Get all resources."""
        pass

    @abc.abstractmethod
    def post(self, resource):
        """Create a resource."""
        pass

    @abc.abstractmethod
    def put(self, resource):
        """Update a resource."""
        pass

    @abc.abstractmethod
    def delete(self, resource_id):
        """Delete a single resource."""
        pass



# Copyright 2014 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add service name to ComputeHost

Revision ID: 10e34bba18e8
Revises: 0_1
Create Date: 2014-04-04 11:00:57.542857

"""

# revision identifiers, used by Alembic.
revision = '10e34bba18e8'
down_revision = '0_1'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('computehosts', sa.Column(
        'service_name', sa.String(length=255), nullable=True))


def downgrade():
    engine = op.get_bind().engine
    if engine.name == 'sqlite':
        # Only for testing purposes with sqlite
        op.execute('CREATE TABLE tmp_computehosts as SELECT id, '
                   'vcpus, cpu_info, hypervisor_type, '
                   'hypervisor_version, hypervisor_hostname, memory_mb, '
                   'local_gb, status '
                   'FROM computehosts')
        op.execute('DROP TABLE computehosts')
        op.execute('ALTER TABLE tmp_computehosts RENAME TO computehosts')
        return

    op.drop_column('computehosts', 'service_name')



# Copyright (c) 2013 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from blazar.manager.oshosts import rpcapi as manager_rpcapi
from blazar import policy
from blazar.utils import trusts


class API(object):
    def __init__(self):
        self.manager_rpcapi = manager_rpcapi.ManagerRPCAPI()

    @policy.authorize('oshosts', 'get')
    def get_computehosts(self, query):
        """List all existing computehosts."""
        return self.manager_rpcapi.list_computehosts(query=query)

    @policy.authorize('oshosts', 'post')
    @trusts.use_trust_auth()
    def create_computehost(self, data):
        """Create new computehost.

        :param data: New computehost characteristics.
        :type data: dict
        """

        return self.manager_rpcapi.create_computehost(data)

    @policy.authorize('oshosts', 'get')
    def get_computehost(self, host_id):
        """Get computehost by its ID.

        :param host_id: ID of the computehost in Blazar DB.
        :type host_id: str
        """
        return self.manager_rpcapi.get_computehost(host_id)

    @policy.authorize('oshosts', 'put')
    def update_computehost(self, host_id, data):
        """Update computehost. Only name changing may be proceeded.

        :param host_id: ID of the computehost in Blazar DB.
        :type host_id: str
        :param data: New computehost characteristics.
        :type data: dict
        """
        return self.manager_rpcapi.update_computehost(host_id, data)

    @policy.authorize('oshosts', 'delete')
    def delete_computehost(self, host_id):
        """Delete specified computehost.

        :param host_id: ID of the computehost in Blazar DB.
        :type host_id: str
        """
        self.manager_rpcapi.delete_computehost(host_id)

    @policy.authorize('oshosts', 'get_allocations')
    def list_allocations(self, query):
        """List all allocations on all computehosts.

        :param query: parameters to query allocations
        :type query: dict
        """
        return self.manager_rpcapi.list_allocations(query)

    @policy.authorize('oshosts', 'get_allocations')
    def get_allocations(self, host_id, query):
        """List all allocations on a specified computehost.

        :param host_id: ID of the computehost in Blazar DB.
        :type host_id: str
        :param query: parameters to query allocations
        :type query: dict
        """
        return self.manager_rpcapi.get_allocations(host_id, query)

    @policy.authorize('oshosts', 'get_resource_properties')
    def list_resource_properties(self, query):
        """List resource properties for hosts."""
        return self.manager_rpcapi.list_resource_properties(query)

    @policy.authorize('oshosts', 'update_resource_properties')
    def update_resource_property(self, property_name, data):
        """Update a host resource property."""
        return self.manager_rpcapi.update_resource_property(
            property_name, data)
